# -*- coding: utf-8 -*-

r"""Convolutional gridding with W projection.

Gridding
--------
The GPU approach is based on [Rom12]_. Each workgroup (thread block) handles a
contiguous range of visibilities, with the threads cooperating to grid each
visibility.  The grid is divided into *bins*. The size of a bin is the size of
the convolution kernel, plus some padding. These bins are further divided into
*tiles*, which are divided into *blocks*. A workitem (thread) handles all
blocks that are at the same offset within their bin. Similarly, a workgroup
handles all tiles that are at the same offset within their bin. At present,
the blocks, tiles and bins are all power-of-two sizes.

For each visibility, the kernel footprint is found, and a bin-sized,
block-aligned bounding box is placed around it. The requirement to be able to
block-align this rectangle means that bins may need to be slightly larger than
the kernel. Every grid point into this box will be updated for this
visibility, and hence the kernel function needs zero padding (on all sides) of
the same amount as the bin padding.

.. tikz:: Mapping of workitems and workgroups to grid points. One workgroup
   contributes to all the green cells. One workitem contributes to all the
   cells marked with a red dot. Here blocks are 2×2, tiles are 4×4 and bins
   are 8×8. The dashed lines show a kernel footprint, and the blue box shows
   the bounding rectangle.

   [x=0.5cm, y=0.5cm]
   \foreach \i in {4, 12}
       \foreach \j in {0, 8}
       {
           \fill[shift={(\i,\j)},green!50!white] (0, 0) rectangle (4, 4);
       }
   \foreach \i in {4,5, 12, 13}
       \foreach \j in {2,3, 10,11}
       {
           \node[shift={(\i,\j)},circle,inner sep=0pt, minimum size=0.2cm,fill=red]
               at (0.5, 0.5) {};
       }
   \draw[help lines] (0, 0) grid[step=0.5cm] (16, 16);
   \draw[very thick] (0, 0) grid[step=4cm] (16, 16);
   \draw[very thick,blue] (2, 2) rectangle (10, 10);
   \draw[very thick,dashed] (2, 3) rectangle (9, 10);

This means that each visibility is processed by multiple work-groups.
Specifically, the number of work-groups loading a visibility is the bin size
over the tile size.

Each thread maintains a number of accumulators: one per position within a
block, per polarization. When moving to the next visibility, any counters that
now correspond to a grid element outside the bounding box are flushed to global
memory (with an atomic add), and reset to a zero value at the new position
that falls inside the bounding box. Provided that the bounding box only moves
slowly, this reduces memory traffic. Note that because the bounding box is
block-aligned, each block is written all at the same time. This allows a lot
of coordinate calculations to be done per-block instead of per-cell, which
improves performance.

Visibilities are loaded in *batches*, whose size equals the number of
workitems in a workgroup. First, all workitems cooperatively load a batch and
store the values in shared memory. Then, each workitem iterates over the
visibilities in the batch. This amortises the costs of the global memory loads
and address computations.

Degridding
----------
Degridding is done similarly to gridding, with a few significant changes:

- Each workgroup is split into "subgroups", each of which functions
  independently as if it were a workgroup. This is done to allow for good
  performance even when the subgroup size is below the minimum workgroup size
  to achieve full occupancy on NVIDIA GPUs (this is not an issue on AMD GCN
  GPUs, which can achieve full occupancy a wavefront per workgroup). Smaller
  subgroups reduce the cost of reductions, particular if they are warp-sized
  and can use warp shuffle instructions. On the other hand, the smaller the
  subgroup, the fewer visibilities can be loaded in each batch.
- Each subgroup loops over the tiles in a bin, instead of the tiles being
  split into different subgroups. This allows each visibility to be completely
  calculated in local memory with no atomics, rather than accumulated in
  global memory. On the other hand, it reduces parallelism. Note, this
  variation is also being considered for gridding.

Weights
-------
There are two sources of weighting: statistical weights and imaging weights
(the latter combining density and taper weights) [Bri95]_. The statistical
weights are pre-multiplied into the visibilities before gridding (see
:py:mod:`.preprocess`), and are handled outside this module for degridding. The
imaging weights are stored in a grid and multiplied by the visibilities as
they are loaded, and have no effect on degridding.

Future planned changes
----------------------
There are two possible ways to handle tiles. The current implementation puts
each tile position into a different workgroup. This has the advantage that it
is not necessary to flush all the accumulators at the end of each batch, but
the disadvantage that every visibility is loaded from global memory multiple
times. A possible alternative (for which further analysis would be needed) is
to load a batch into shared memory, and then process multiple tiles in series
inside the workgroup, reusing the values in shared memory each time.

Another reason to use separate workgroups for each tile is that it may be
possible to eliminate atomic operations on small devices (such as a Tegra K1
or X1) by having each workgroup process *all* the visibilities, rather than
splitting them up to create more workgroups. This eliminates the potential for
race conditions, because each workgroup is handling a disjoint part of the
grid. This does limit the amount of available parallelism to the number of
blocks per bin, which is why it is probably only practical on embedded
devices, and even then may limit the block size.

.. [Rom12] Romein, John W. 2012. An efficient work-distribution strategy for
   gridding radio-telescope data on GPUs. In *Proceedings of the 26th ACM International
   Conference on Supercomputing (ICS '12)*, 321-330.
   http://www.astron.nl/~romein/papers/ICS-12/gridding.pdf

.. include:: macros.rst
"""

import math
import logging

import numpy as np
import pkg_resources

import katsdpsigproc.accel as accel
import katsdpsigproc.tune as tune

import katsdpimager.types
from . import numba


logger = logging.getLogger(__name__)


def kaiser_bessel(x, width, beta):
    r"""Evaluate Kaiser-Bessel window function. Refer to
    http://www.dsprelated.com/freebooks/sasp/Kaiser_Window.html
    for details.

    Parameters
    ----------
    x : array-like, float
        Sample positions
    width : float
        The kernel has support :math:`[-\frac{1}{2}W, \frac{1}{2}W]`
    beta : float
        Shape parameter
    """
    param = 1 - (2 * x / width)**2
    # The np.maximum is to protect against runtime warnings for taking
    # sqrt of negative values. The actual values in this situation are
    # irrelevent due to the np.select call.
    values = np.i0(beta * np.sqrt(np.maximum(0, param))) / np.i0(beta)
    return np.select([param >= 0], [values])


def kaiser_bessel_fourier(f, width, beta, out=None):
    r"""
    Evaluate the continuous Fourier transform of :func:`kaiser_bessel`.
    Note that since the function is even and real, this is also the inverse
    Fourier transform.

    Parameters
    ----------
    f : array-like, float
        Sample positions (frequency)
    width : float
        The kernel has support :math:`[-\frac{1}{2}W, \frac{1}{2}W]`
    beta : float
        Shape parameter
    out : array-like, float
        If specified, result is written into it
    """
    alpha = beta / math.pi
    # np.lib.scimath.sqrt returns complex values for negative inputs, whereas
    # np.sqrt returns NaN. We take the real component because np.sinc returns
    # complex if it has a complex input, even though the imaginary part is 0.
    ans = width / np.i0(beta) * np.sinc(np.lib.scimath.sqrt((width * f)**2 - alpha * alpha)).real
    if out is not None:
        out[:] = ans
        return out
    else:
        return ans


def antialias_kernel(width, oversample, beta=None):
    r"""Generate 1D anti-aliasing kernel. The return value is 2-dimensional.
    The first dimension selects the subpixel position, the second
    selects the pixel position.

    The returned kernels all have the same size, but it is not necessarily
    the same size as `width` (in fact, `width` may be non-integer). Given
    a real-valued :math:`x` coordinate and a grid point :math:`x_g` to be
    updated, the indices are computed as follows:

    .. math::

       x_0 &= \lfloor x\rfloor\\
       s &= \lfloor (x - x_0)\cdot\text{oversample}\rfloor\\
       u &= x_g - x_0 + \tfrac12 \lvert \text{kernel}\rvert - 1

    Then `s` is the subpixel index and `u` is the pixel index.

    TODO: normalisation so that integral is 1?

    Parameters
    ----------
    width : float
        Kernel support
    oversample : int
        Number of samples per unit step
    beta : float, optional
        Shape parameter for Kaiser-Bessel window
    """
    if beta is None:
        # Puts the first null of the taper function at the edge of the image
        beta = math.pi * math.sqrt(0.25 * width**2 - 1.0)
        # Move the null outside the image, to avoid numerical instabilities.
        # This will cause a small amount of aliasing at the edges, which
        # ideally should be handled by clipping the image.
        beta *= 1.2
    hsize = int(math.ceil(0.5 * width))
    size = 2 * hsize
    # The kernel only contains real values, but we store complex so that we
    # are ready for W projection.
    kernel = np.empty((oversample, size), np.complex64)
    for s in range(oversample):
        bias = (s + 0.5) / oversample + hsize - 1
        kernel[s, :] = kaiser_bessel(np.arange(size) - bias, width, beta)
    return kernel


def antialias_w_kernel(
        cell_wavelengths, w, width,
        oversample, antialias_width, image_oversample, beta,
        out=None):
    r"""Computes a combined anti-aliasing and W-projection kernel.

    The format of the returned kernel is similar to :func:`antialias_kernel`.
    In particular, only a 1D kernel is returned.  Note that while the W kernel
    is not truly separable, the small-angle approximation
    :math:`\sqrt{1-l^2-m^2}-1 \approx -\frac{1}{2}(l^2+m^2)-\frac{5}{24}(l^4+m^4)`
    makes it very close to separable [#]_.

    .. [#] The error in the approximation above is on the order of :math:`5\times 10^{-6}`
       for a 10 degree field of view, and less than :math:`10^{-8}` for a 2 degree FOV.

    We multiply the inverse Fourier transform of the (idealised) antialiasing
    kernel with the W correction in image space, to obtain a closed-form
    function we can evaluate in image space. This is then Fourier-transformed
    to UV space. To reduce aliasing, the image-space function is oversampled by
    a factor of `image_oversample`, and the transform result is then truncated.

    All the continuous functions that we are approximating in UV space are
    expressed as functions of the number of wavelengths.

    The output kernel is sampled at the centres of bins, which are at
    half-subpixel offsets. To create this shift in UV space, we multiply by the
    appropriate complex exponential in image space.

    Parameters
    ----------
    cell_wavelengths : float
        Size of a UV cell in wavelengths
    w : float
        w component of baseline, in wavelengths
    width : int
        Support of combined kernel, in cells
    oversample : int
        Number of samples per unit step in UV space
    antialias_width : float
        Support for anti-alias kernel, in cells
    image_oversample : int
        Oversampling factor in image space. Larger values reduce aliasing in
        the output and increase time and memory required for computation, but
        do not affect the size of the resulting kernel.
    beta : float, optional
        Shape parameter for Kaiser-Bessel window
    out : array-like, optional
        If specified, the output is returned in this array
    """
    def image_func(l):
        # The kaiser_bessel function is designed around units of cells rather
        # than wavelengths. We thus want the Fourier transform of
        # kaiser_bessel(u / cell_wavelengths).
        scale_l = l * cell_wavelengths
        aa_factor = cell_wavelengths * kaiser_bessel_fourier(scale_l, antialias_width, beta)
        shift_arg = shift_by * l
        l2 = l * l
        l4 = l2 * l2
        w_arg = -w * (-0.5 * l2 - 5.0/24.0 * l4)
        return aa_factor * np.exp(2j * math.pi * (w_arg + shift_arg))

    out_pixels = oversample * width
    assert out_pixels % 2 == 0, "Odd number of pixels is not tested"
    pixels = out_pixels * image_oversample
    # Convert uv-space width to wavelengths
    uv_width = width * cell_wavelengths * image_oversample
    # Compute other support and step sizes
    image_step = 1 / uv_width
    # Determine sample points in image space
    l = (np.arange(pixels) - (pixels // 2)) * image_step
    # Evaluate function in image space
    shift_by = -0.5 * cell_wavelengths / oversample
    image_values = image_func(l)
    # Convert to UV space. The multiplication is because we're using a DFT to
    # approximate a continuous FFT.
    uv_values = np.fft.fft(np.fft.ifftshift(image_values)) * image_step
    # Crop to area of interest, and swap halves to put DC in the middle
    uv_values = np.concatenate((uv_values[-(out_pixels // 2):], uv_values[:(out_pixels // 2)]))
    # Split up into subkernels. Since the subpixel index indicates the subpixel
    # position of the visibility, rather than the kernel tap, it runs backwards
    # in the kernel indexing.
    kernel = np.reshape(uv_values, (oversample, width), order='F')[::-1, :]
    # Convert to C memory layout
    if out is None:
        out = np.empty_like(kernel)
    out[:] = kernel
    return out


@numba.jit(nopython=True)
def subpixel_coord(x, oversample):
    """Return pixel and subpixel index, as described in :func:`antialias_kernel`"""
    xs = int(np.floor(x * oversample))
    return xs // oversample, xs % oversample


class ConvolutionKernel:
    """Separable convolution kernel with metadata. The kernel combines anti-aliasing
    with W correction.

    Parameters
    ----------
    image_parameters : :class:`katsdpimager.parameters.ImageParameters`
        Command-line parameters with image properties
    grid_parameters : :class:`katsdpimager.parameters.GridParameters`
        Gridding parameters
    data : array-like
        If specified, used as a backing data store. Otherwise, a plain numpy
        array is used.
    """
    def __init__(self, image_parameters, grid_parameters, data=None):
        self.grid_parameters = grid_parameters
        shape = (grid_parameters.w_planes, grid_parameters.oversample, grid_parameters.kernel_width)
        if data is None:
            self.data = np.empty(shape, np.complex64)
        else:
            self.data = data
        cell_wavelengths = float(image_parameters.cell_size / image_parameters.wavelength)
        # Separation in w between slices
        w_slice_wavelengths = float(grid_parameters.max_w
                                    / (grid_parameters.w_slices * image_parameters.wavelength))
        # Separation in w between planes
        w_plane_wavelengths = w_slice_wavelengths / grid_parameters.w_planes
        # Puts the first null of the taper function at the edge of the image
        self.beta = math.pi * math.sqrt(0.25 * grid_parameters.antialias_width**2 - 1.0)
        # Move the null outside the image, to avoid numerical instabilities.
        # This will cause a small amount of aliasing at the edges, which
        # ideally should be handled by clipping the image.
        self.beta *= 1.2
        # TODO: use sqrt(w) scaling as in Cornwell, Golap and Bhatnagar (2008)?
        # TODO: can halve work and memory by exploiting conjugate symmetry
        # Find w for the midpoint of the final plane
        max_w_wavelengths = (w_slice_wavelengths - w_plane_wavelengths) * 0.5
        ws = np.linspace(-max_w_wavelengths, max_w_wavelengths, grid_parameters.w_planes)
        for i, w in enumerate(ws):
            antialias_w_kernel(
                cell_wavelengths, w, grid_parameters.kernel_width,
                grid_parameters.oversample,
                grid_parameters.antialias_width,
                grid_parameters.image_oversample,
                self.beta, out=self.data[i, ...])

    @classmethod
    def get_bin_size(cls, grid_parameters, tile_x, tile_y, pad):
        """Determine appropriate bin size given alignment restrictions."""
        size = max(tile_x, tile_y)
        # Round up to a power of 2
        while size < grid_parameters.kernel_width + pad:
            size *= 2
        if size != grid_parameters.kernel_width + pad:
            logger.info("kernel size rounded up to %d", size)
        assert size % tile_x == 0
        assert size % tile_y == 0
        return size

    def taper(self, N, out=None):
        """Return the Fourier transform of the antialiasing kernel for
        an N-pixel 1D image.

        The combined filter is sampled and then applied as a piecewise constant
        function, which corresponds to convolution with a rect. In image space,
        this corresponds to aliasing, then multiplication by a sinc. We need to
        include the sinc in the Fourier transform.

        Parameters
        ----------
        N : int
            Number of pixels in the image
        out : array-like, optional
            If provided, is used to store the result
        """
        x = np.arange(N) / N - 0.5
        out = kaiser_bessel_fourier(x, self.grid_parameters.antialias_width, self.beta, out)
        out *= np.sinc(x / self.grid_parameters.oversample)
        return out


class ConvolutionKernelDevice(ConvolutionKernel):
    """A :class:`ConvolutionKernel` that stores data in a device array."""
    def __init__(self, context, image_parameters, grid_parameters, pad=0, allocator=None):
        if allocator is None:
            allocator = accel.DeviceAllocator(context)
        out = allocator.allocate(
            (grid_parameters.w_planes,
             grid_parameters.oversample,
             grid_parameters.kernel_width + 2 * pad),
            np.complex64)
        if isinstance(out, accel.SVMArray):
            host = out
        else:
            host = out.empty_like()
        host.fill(0)
        super().__init__(
            image_parameters, grid_parameters, host[:, :, pad:grid_parameters.kernel_width + pad])
        if not isinstance(out, accel.SVMArray):
            queue = context.create_command_queue()
            out.set(queue, host)
        self.padded_data = out
        self.pad = pad

    @property
    def bin_size(self):
        return self.data.shape[-1] + self.pad

    def parameters(self):
        """Parameters for templating CUDA/OpenCL kernels"""
        return {
            'convolve_kernel_slice_stride':
                self.padded_data.padded_shape[2],
            'convolve_kernel_w_stride': np.product(self.padded_data.padded_shape[1:]),
            'bin_x': self.bin_size,
            'bin_y': self.bin_size,
        }


def _autotune_uv(context, pixels, bin_size, oversample):
    """Creates UV test data for autotuning gridding and degridding.

    The visibilities and weights are irrelevant, but the UV coordinates are
    important because the performance is heavily data-dependent. Compression
    ensures that samples in the same subgrid cell are compressed, but each
    sample is typically close to the previous one, moving along an elliptic
    track. Rather than go to the full complexity of modelling such tracks, we
    use a number of linear tracks at a variety of slopes.

    The return value has the format required by the *uv* slot in
    :class:`Gridder`.
    """
    # Set bounds of the UV coordinates in pixels. This is a slightly
    # conservative bound.
    low = bin_size
    high = pixels - bin_size
    track_length = (high - low) * oversample
    tracks = 256 * 1024 // track_length
    out = np.empty((tracks, track_length, 4), np.int16)
    for i in range(tracks):
        angle = 2 * math.pi * i / tracks
        dx = math.cos(angle)
        dy = math.sin(angle)
        scale = max(abs(dx), abs(dy))
        # step is either (1, slope), (-1, slope), (slope, 1) or (slope, -1)
        step = np.array([dx / scale, dy / scale])
        indices = np.arange(track_length) - track_length / 2
        sub_uv = np.outer(indices, step)
        sub_uv = np.round(sub_uv).astype(np.int32)
        out[i, :, 0:2] = sub_uv // oversample
        out[i, :, 2:4] = sub_uv % oversample
    # Flatten the separate tracks into one dimension
    return out.reshape((-1, 4))


def _autotune_arrays(command_queue, oversample, real_dtype, num_polarizations, bin_size, pad):
    """Creates device arrays for autotuning gridding and degridding. The
    UV coordinates are provided by :func:`_autotune_uv`, while the other
    arrays are zeroed out.

    Parameters
    ----------
    command_queue : |CommandQueue|
        Command queue for zeroing the arrays
    oversample : int
        Grid oversampling
    real_dtype : {np.float32, np.complex64}
        Floating-point type for grid value
    num_polarizations : int
        Number of polarizations in grid
    bin_size : int
        Bin size
    pad : int
        Amount of padding around edges of convolution kernel to allow for
        blocking. This must be at greater than or equal to the maximum
        block size (X or Y) minus one.

    Returns
    -------
    grid,weights_grid,uv,w_plane,vis,convolve_kernel : :class:`katsdpsigproc.accel.DeviceArray`
        Device arrays to be passed to the kernels
    """
    context = command_queue.context
    pixels = 256
    complex_dtype = katsdpimager.types.real_to_complex(real_dtype)
    uv_host = _autotune_uv(context, pixels, bin_size, oversample)
    num_vis = len(uv_host)
    uv = accel.DeviceArray(context, uv_host.shape, np.int16)
    grid = accel.DeviceArray(context, (num_polarizations, pixels, pixels), complex_dtype)
    weights_grid = accel.DeviceArray(context, (num_polarizations, pixels, pixels), np.float32)
    w_plane = accel.DeviceArray(context, (num_vis,), np.int16)
    vis = accel.DeviceArray(context, (num_vis, num_polarizations), np.complex64)
    convolve_kernel = accel.DeviceArray(context, (1, oversample, bin_size + pad), np.complex64)
    uv.set(command_queue, uv_host)
    grid.zero(command_queue)
    weights_grid.zero(command_queue)
    w_plane.zero(command_queue)
    vis.zero(command_queue)
    convolve_kernel.zero(command_queue)
    return grid, weights_grid, uv, w_plane, vis, convolve_kernel


class GridderTemplate:
    autotune_version = 6

    def __init__(self, context, image_parameters, grid_parameters, tuning=None):
        if tuning is None:
            tuning = self.autotune(
                context, grid_parameters.oversample, image_parameters.real_dtype,
                len(image_parameters.polarizations))
        self.grid_parameters = grid_parameters
        self.image_parameters = image_parameters
        self.wgs_x = tuning['wgs_x']
        self.wgs_y = tuning['wgs_y']
        self.multi_x = tuning['multi_x']
        self.multi_y = tuning['multi_y']
        min_pad = max(self.multi_x, self.multi_y) - 1
        self.tile_x = self.wgs_x * self.multi_x
        self.tile_y = self.wgs_y * self.multi_y
        bin_size = ConvolutionKernel.get_bin_size(
            grid_parameters, self.tile_x, self.tile_y, min_pad)
        pad = bin_size - grid_parameters.kernel_width
        self.convolve_kernel = ConvolutionKernelDevice(
            context, image_parameters, grid_parameters, pad)
        parameters = {
            'real_type': katsdpimager.types.dtype_to_ctype(image_parameters.real_dtype),
            'multi_x': self.multi_x,
            'multi_y': self.multi_y,
            'wgs_x': self.wgs_x,
            'wgs_y': self.wgs_y,
            'num_polarizations': len(image_parameters.polarizations)
        }
        parameters.update(self.convolve_kernel.parameters())
        self.program = accel.build(
            context, "imager_kernels/grid.mako", parameters,
            extra_dirs=[pkg_resources.resource_filename(__name__, '')])

    @classmethod
    @tune.autotuner(test={'wgs_x': 16, 'wgs_y': 8, 'multi_x': 2, 'multi_y': 2})
    def autotune(cls, context, oversample, real_dtype, num_polarizations):
        # This autotuning code is unusually complex because it tries to avoid
        # actually creating the convolution kernels. Doing so is slow, would
        # require fake values of many parameters to be carefully chosen, and
        # could potentially lead to different bin sizes for different tuning
        # parameters. Instead, we directly build the program and data
        # structures.
        queue = context.create_tuning_command_queue()
        bin_size = 32
        pad = 3
        grid, weights_grid, uv, w_plane, vis, convolve_kernel = _autotune_arrays(
            queue, oversample, real_dtype, num_polarizations, bin_size, pad)
        num_vis = uv.shape[0]

        def generate(multi_x, multi_y, wgs_x, wgs_y):
            assert multi_x <= pad + 1 and multi_y <= pad + 1
            # No point having less than a warp per workgroup
            if wgs_x * wgs_y < context.device.simd_group_size:
                return None
            # Only the total size really matters, so it is not necessary
            # to try all possible shapes. For block shape, it is important
            # to be as square as possible.
            if wgs_x != wgs_y and wgs_y != wgs_x * 2:
                return None
            if multi_x != multi_y and multi_x != multi_y * 2:
                return None
            # This can cause launch timeouts due to excessive register pressure
            # and spilling.
            if multi_x * multi_y * num_polarizations * real_dtype.itemsize > 256:
                return None
            tile_x = wgs_x * multi_x
            tile_y = wgs_y * multi_y
            if tile_x > bin_size or tile_y > bin_size:
                return None
            parameters = {
                'real_type': katsdpimager.types.dtype_to_ctype(real_dtype),
                'multi_x': multi_x,
                'multi_y': multi_y,
                'wgs_x': wgs_x,
                'wgs_y': wgs_y,
                'bin_x': bin_size,
                'bin_y': bin_size,
                'num_polarizations': num_polarizations,
                'convolve_kernel_slice_stride': convolve_kernel.padded_shape[2],
                'convolve_kernel_w_stride': np.product(convolve_kernel.padded_shape[1:])
            }
            program = accel.build(
                context, "imager_kernels/grid.mako", parameters,
                extra_dirs=[pkg_resources.resource_filename(__name__, '')])
            kernel = program.get_kernel('grid')
            uv_bias = (bin_size - 1) // 2 - grid.shape[-1] // 2

            def fn():
                Gridder.static_run(
                    queue, kernel, wgs_x, wgs_y, tile_x, tile_y, bin_size,
                    num_vis, uv_bias,
                    grid, weights_grid, uv, w_plane, vis, convolve_kernel)
            return tune.make_measure(queue, fn)

        return tune.autotune(
            generate,
            multi_x=[1, 2, 4],
            multi_y=[1, 2, 4],
            wgs_x=[4, 8, 16],
            wgs_y=[4, 8, 16])

    def instantiate(self, *args, **kwargs):
        return Gridder(self, *args, **kwargs)


class VisOperation(accel.Operation):
    """Base for operation classes that store visibility data in GPU buffers.

    .. rubric:: Slots

    **uv** : array of int16×4
        The first two elements for each visibility are the
        UV coordinates of the first grid cell to be updated. The other two are
        the subpixel U and V coordinates.
    **w_plane** : array of int16
        W plane index per visibility, clamped to the range of allocated w planes
    **vis** : array of complex64 × pols
        Visibilities, which are pre-multiplied by statistical weights. For
        imaging these are inputs.  For prediction these are visibilities on
        input and residual visibilities on output.

    Parameters
    ----------
    command_queue : |CommandQueue|
        Command queue for the operation
    num_polarizations : int
        Size of vis buffer on polarization dimension
    max_vis : int
        Number of visibilities that can be supported per kernel invocation
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots
    """
    def __init__(self, command_queue, num_polarizations, max_vis, allocator=None):
        super().__init__(command_queue, allocator)
        self.max_vis = max_vis
        self.slots['uv'] = accel.IOSlot(
            (max_vis, accel.Dimension(4, exact=True)), np.int16)
        self.slots['w_plane'] = accel.IOSlot((max_vis,), np.int16)
        self.slots['vis'] = accel.IOSlot(
            (max_vis, accel.Dimension(num_polarizations, exact=True)), np.complex64)
        self._num_vis = 0

    @property
    def num_vis(self):
        return self._num_vis

    @num_vis.setter
    def num_vis(self, n):
        """Change the number of actual visibilities stored in the buffers."""
        if n < 0 or n > self.max_vis:
            raise ValueError('Number of visibilities {} is out of range 0..{}'.format(
                n, self.max_vis))
        self._num_vis = n

    def set_coordinates(self, uv, sub_uv, w_plane):
        """Set the UVW coordinates.

        Before calling, first set :attr:`num_vis`.
        """
        N = self.num_vis
        if len(uv) != N or len(sub_uv) != N or len(w_plane) != N:
            raise ValueError('Lengths do not match')
        self.buffer('uv')[:N, 0:2] = uv
        self.buffer('uv')[:N, 2:4] = sub_uv
        self.buffer('w_plane')[:N] = w_plane

    def set_vis(self, vis):
        """Set input visibilities.

        Before calling, first set :attr:`num_vis`.
        """
        N = self.num_vis
        if len(vis) != N:
            raise ValueError('Lengths do not match')
        self.buffer('vis')[:N] = vis


class GridDegrid(VisOperation):
    """Base class for :class:`Gridder` and :class:`Degridder`.

    .. rubric:: Slots

    In addition to those in :class:`VisOperation`:

    **grid** : array of pols × height × width, complex
        Grid (output for :class:`Gridder`, input for :class:`Degridder`)
    **weights_grid** : array of pols × height × width, float32
        Grid of density weights (input for :class:`Gridder`, absent for :class:`Degridder`)
    **weights** : array of float32 × pols
        Statistical weights for visibilities (:class:`Degridder` only).

    Parameters
    ----------
    template : :class:`GridderTemplate` or :class:`DegridderTemplate`
        Operation template
    command_queue : |CommandQueue|
        Command queue for the operation
    array_parameters : :class:`~katsdpimager.parameters.ArrayParameters`
        Array parameters
    max_vis : int
        Number of visibilities that can be supported per kernel invocation
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots

    Raises
    ------
    ValueError
        If the longest baseline is too large to find within the grid
    """
    def __init__(self, template, command_queue, array_parameters,
                 max_vis, allocator=None):
        num_polarizations = len(template.image_parameters.polarizations)
        super().__init__(command_queue, num_polarizations, max_vis, allocator)
        # Check that longest baseline won't cause an out-of-bounds access
        max_uv_src = float(array_parameters.longest_baseline / template.image_parameters.cell_size)
        convolve_kernel_size = template.convolve_kernel.padded_data.shape[-1]
        # I don't think the + 1 is actually needed, but it's a safety factor in
        # case I've made an off-by-one error in the maths.
        grid_pixels = 2 * (int(max_uv_src) + convolve_kernel_size // 2 + 1)
        if grid_pixels > template.image_parameters.pixels:
            raise ValueError('image_oversample is too small '
                             'to capture all visibilities in the UV plane')
        self.template = template
        self.slots['grid'] = accel.IOSlot(
            (num_polarizations, grid_pixels, grid_pixels),
            template.image_parameters.complex_dtype)

    def parameters(self):
        return {
            'grid_parameters': self.template.grid_parameters,
            'image_parameters': self.template.image_parameters
        }


class Gridder(GridDegrid):
    """Instantiation of :class:`GridderTemplate`. See :class:`GridDegrid` for
    details.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.slots['weights_grid'] = accel.IOSlot(self.slots['grid'].shape, np.float32)
        self._kernel = self.template.program.get_kernel('grid')

    @classmethod
    def static_run(cls, command_queue, kernel,
                   wgs_x, wgs_y, tile_x, tile_y, bin_size, num_vis, uv_bias,
                   grid, weights_grid, uv, w_plane, vis, convolve_kernel):
        """Implementation of :meth:`_run` with all the parameters explicit.
        This is not intended for direct use, but is used for autotuning to
        execute the kernel while bypassing a lot of unwanted setup.

        Parameters
        ----------
        command_queue : |CommandQueue|
            Command queue for the operation
        kernel : :class:`katsdpsigproc.cuda.Kernel` or :class:`katsdpsigproc.opencl.Kernel`
            Compiled kernel to run
        wgs_x,wgs_y,tile_x,tile_y,bin_size : int
            Tuning parameters (see :class:`GridTemplate`)
        num_vis : int
            Number of visibilities to grid
        uv_bias : int
            Bias for UV coordinates to center the kernel
        grid,weights_grid,uv,w_plane,vis,convolve_kernel : \
                :class:`~katsdpsigproc.accel.DeviceArray`-like
            Data passed to the kernel
        """
        if num_vis == 0:
            return
        tiles_x = bin_size // tile_x
        tiles_y = bin_size // tile_y
        vis_per_workgroup = 1024
        workgroups_z = accel.divup(num_vis, vis_per_workgroup)
        # Recompute vis_per_workgroup: a smaller value may now be possible
        # for the given value of workgroups_z.
        vis_per_workgroup = accel.divup(num_vis, workgroups_z)
        # Make sure that vis_per_workgroup is a multiple of the batch size.
        # If it is not, then every workgroup will finish with a partial batch,
        # rather than only the last workgroup.
        vis_per_workgroup = accel.roundup(vis_per_workgroup, wgs_x * wgs_y)
        half_u = weights_grid.shape[1] // 2
        half_v = weights_grid.shape[2] // 2
        weights_address_bias = half_v * weights_grid.padded_shape[2] + half_u
        command_queue.enqueue_kernel(
            kernel,
            [
                grid.buffer,
                np.int32(grid.padded_shape[2]),
                np.int32(grid.padded_shape[1] * grid.padded_shape[2]),
                weights_grid.buffer,
                np.int32(weights_grid.padded_shape[2]),
                np.int32(weights_grid.padded_shape[1] * weights_grid.padded_shape[2]),
                uv.buffer,
                w_plane.buffer,
                vis.buffer,
                convolve_kernel.buffer,
                np.int32(uv_bias),
                np.int32(weights_address_bias),
                np.int32(vis_per_workgroup),
                np.int32(num_vis)
            ],
            global_size=(wgs_x * tiles_x,
                         wgs_y * tiles_y,
                         workgroups_z),
            local_size=(wgs_x, wgs_y, 1)
        )

    def _run(self):
        kernel_width = self.template.grid_parameters.kernel_width
        uv_bias = ((kernel_width - 1) // 2 + self.template.convolve_kernel.pad
                   - self.slots['grid'].shape[-1] // 2)
        self.static_run(
            self.command_queue, self._kernel,
            self.template.wgs_x, self.template.wgs_y,
            self.template.tile_x, self.template.tile_y,
            self.template.convolve_kernel.bin_size,
            self.num_vis,
            uv_bias,
            self.buffer('grid'),
            self.buffer('weights_grid'),
            self.buffer('uv'),
            self.buffer('w_plane'),
            self.buffer('vis'),
            self.template.convolve_kernel.padded_data)


class DegridderTemplate:
    autotune_version = 3

    def __init__(self, context, image_parameters, grid_parameters, tuning=None):
        if tuning is None:
            tuning = self.autotune(
                context, grid_parameters.oversample, image_parameters.real_dtype,
                len(image_parameters.polarizations))
        self.grid_parameters = grid_parameters
        self.image_parameters = image_parameters
        self.wgs_x = tuning['wgs_x']
        self.wgs_y = tuning['wgs_y']
        self.wgs_z = tuning['wgs_z']
        self.multi_x = tuning['multi_x']
        self.multi_y = tuning['multi_y']
        self.tile_x = self.wgs_x * self.multi_x
        self.tile_y = self.wgs_y * self.multi_y
        min_pad = max(self.multi_x, self.multi_y) - 1
        bin_size = ConvolutionKernel.get_bin_size(grid_parameters, self.tile_x, self.tile_y,
                                                  min_pad)
        pad = bin_size - grid_parameters.kernel_width
        # Note: we can't necessarily use the same kernel as for gridding,
        # because different tuning parameters will affect the kernel padding.
        # TODO: should still reuse the work done to compute the function, and
        # simply adjust the padding.
        self.convolve_kernel = ConvolutionKernelDevice(
            context, image_parameters, grid_parameters, pad)
        parameters = {
            'real_type': katsdpimager.types.dtype_to_ctype(image_parameters.real_dtype),
            'multi_x': self.multi_x,
            'multi_y': self.multi_y,
            'wgs_x': self.wgs_x,
            'wgs_y': self.wgs_y,
            'wgs_z': self.wgs_z,
            'num_polarizations': len(image_parameters.polarizations)
        }
        parameters.update(self.convolve_kernel.parameters())
        self.program = accel.build(
            context, "imager_kernels/degrid.mako", parameters,
            extra_dirs=[pkg_resources.resource_filename(__name__, '')])

    @classmethod
    @tune.autotuner(test={'wgs_x': 8, 'wgs_y': 4, 'wgs_z': 4, 'multi_x': 2, 'multi_y': 4})
    def autotune(cls, context, oversample, real_dtype, num_polarizations):
        queue = context.create_tuning_command_queue()
        bin_size = 32
        grid, weights_grid, uv, w_plane, vis, convolve_kernel = _autotune_arrays(
            queue, oversample, real_dtype, num_polarizations, bin_size, 3)
        weights = accel.DeviceArray(context, vis.shape, np.complex64)
        num_vis = uv.shape[0]

        def generate(multi_x, multi_y, wgs_x, wgs_y, wgs_z):
            # No point having less than a warp per workgroup
            if wgs_x * wgs_y * wgs_z < context.device.simd_group_size:
                return None
            # Avoid subgroups bigger than a warp, since they can't use
            # shuffle reduction
            if wgs_x * wgs_y > context.device.simd_group_size:
                return None
            # Cut down number of configurations to test
            if multi_x != multi_y and multi_x != multi_y * 2:
                return None
            if wgs_y != wgs_x and wgs_y != wgs_x * 2:
                return None
            if multi_x * multi_y * num_polarizations * real_dtype.itemsize > 128:
                return None
            # Eliminate configurations that don't fit in a bin
            if wgs_x * multi_x > bin_size or wgs_y * multi_y > bin_size:
                return None
            parameters = {
                'real_type': katsdpimager.types.dtype_to_ctype(real_dtype),
                'multi_x': multi_x,
                'multi_y': multi_y,
                'wgs_x': wgs_x,
                'wgs_y': wgs_y,
                'wgs_z': wgs_z,
                'num_polarizations': num_polarizations,
                'convolve_kernel_slice_stride': convolve_kernel.padded_shape[2],
                'convolve_kernel_w_stride': np.product(convolve_kernel.padded_shape[1:]),
                'bin_x': bin_size,
                'bin_y': bin_size
            }
            program = accel.build(
                context, "imager_kernels/degrid.mako", parameters,
                extra_dirs=[pkg_resources.resource_filename(__name__, '')])
            kernel = program.get_kernel('degrid')
            uv_bias = (bin_size - 1) // 2 - grid.shape[-1] // 2

            def fn():
                Degridder.static_run(
                    queue, kernel, wgs_x, wgs_y, wgs_z,
                    num_vis, uv_bias,
                    grid, uv, w_plane, weights, vis, convolve_kernel)
            return tune.make_measure(queue, fn)

        return tune.autotune(
            generate,
            multi_x=[1, 2, 4],
            multi_y=[1, 2, 4],
            wgs_x=[4, 8],
            wgs_y=[4, 8],
            wgs_z=[1, 2, 4, 8])

    def instantiate(self, *args, **kwargs):
        return Degridder(self, *args, **kwargs)


class Degridder(GridDegrid):
    """Instantiation of :class:`DegridderTemplate`. See :class:`GridDegrid` for
    details.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        num_polarizations = len(self.template.image_parameters.polarizations)
        self.slots['weights'] = accel.IOSlot(
            (self.max_vis, accel.Dimension(num_polarizations, exact=True)), np.float32)
        self._kernel = self.template.program.get_kernel('degrid')

    def set_weights(self, weights):
        """Set statistical weights on visibilities.

        Before calling, set :attr:`num_vis`.
        """
        N = self.num_vis
        if len(weights) != N:
            raise ValueError('Lengths do not match')
        self.buffer('weights')[:N] = weights

    @classmethod
    def static_run(cls, command_queue, kernel,
                   wgs_x, wgs_y, wgs_z, num_vis, uv_bias,
                   grid, uv, w_plane, weights, vis, convolve_kernel):
        if num_vis == 0:
            return
        batch_size = wgs_x * wgs_y
        subgroups = accel.divup(num_vis, batch_size)
        workgroups = accel.divup(subgroups, wgs_z)
        command_queue.enqueue_kernel(
            kernel,
            [
                grid.buffer,
                np.int32(grid.padded_shape[2]),
                np.int32(grid.padded_shape[1] * grid.padded_shape[2]),
                uv.buffer,
                w_plane.buffer,
                weights.buffer,
                vis.buffer,
                convolve_kernel.buffer,
                np.int32(uv_bias),
                np.int32(num_vis)
            ],
            global_size=(wgs_x * workgroups,
                         wgs_y,
                         wgs_z),
            local_size=(wgs_x, wgs_y, wgs_z)
        )

    def _run(self):
        kernel_width = self.template.grid_parameters.kernel_width
        uv_bias = ((kernel_width - 1) // 2 + self.template.convolve_kernel.pad
                   - self.slots['grid'].shape[-1] // 2)
        self.static_run(
            self.command_queue, self._kernel,
            self.template.wgs_x, self.template.wgs_y, self.template.wgs_z,
            self.num_vis,
            uv_bias,
            self.buffer('grid'),
            self.buffer('uv'),
            self.buffer('w_plane'),
            self.buffer('weights'),
            self.buffer('vis'),
            self.template.convolve_kernel.padded_data)


@numba.jit(nopython=True)
def _grid(kernel, grid, weights_grid, uv, sub_uv, w_plane, vis, sample):
    """Internal implementation of :meth:`GridderHost.grid`, split out so that
    Numba can JIT it.
    """
    ksize = kernel.shape[2]
    uv_bias = (ksize - 1) // 2 - grid.shape[2] // 2
    for row in range(uv.shape[0]):
        u0 = uv[row, 0] - uv_bias
        v0 = uv[row, 1] - uv_bias
        sub_u, sub_v = sub_uv[row]
        weights_u = uv[row, 0] + weights_grid.shape[2] // 2
        weights_v = uv[row, 1] + weights_grid.shape[1] // 2
        for i in range(vis.shape[1]):
            sample[i] = vis[row, i] * weights_grid[i, weights_v, weights_u]
        for j in range(ksize):
            for k in range(ksize):
                kernel_sample = kernel[w_plane[row], sub_v, j] * kernel[w_plane[row], sub_u, k]
                weight = np.conj(kernel_sample)
                for pol in range(grid.shape[0]):
                    grid[pol, int(v0 + j), int(u0 + k)] += sample[pol] * weight


class VisOperationHost:
    """Equivalent to :class:`VisOperation` on the host."""
    def __init__(self):
        self._num_vis = 0
        self.uv = None
        self.sub_uv = None
        self.w_plane = None
        self.vis = None

    @property
    def num_vis(self):
        return self._num_vis

    @num_vis.setter
    def num_vis(self, value):
        self._num_vis = value
        self.uv = None
        self.sub_uv = None
        self.w_plane = None
        self.vis = None

    def set_coordinates(self, uv, sub_uv, w_plane):
        """Set UVW coordinates for the visibilities.

        Parameters
        ----------
        uv : 2D array, integer
            Preprocessed grid UV coordinates
        sub_uv : 2D array, integer
            Preprocessed grid UV sub-pixel coordinates
        w_plane : 1D array, integer
            Preprocessed grid W plane coordinates
        """
        N = self.num_vis
        if len(uv) != N or len(sub_uv) != N or len(w_plane) != N:
            raise ValueError('Lengths do not match')
        self.uv = uv
        self.sub_uv = sub_uv
        self.w_plane = w_plane

    def set_vis(self, vis):
        """Set input visibility data.

        vis : 2D ndarray of complex or real
            Visibility data, indexed by sample and polarization, and
            pre-multiplied by all weights
        """
        if len(vis) != self.num_vis:
            raise ValueError('Lengths do not match')
        self.vis = vis


class GridDegridHost(VisOperationHost):
    """Common code shared by :class:`GridderHost` and :class:`DegridderHost`."""
    def __init__(self, image_parameters, grid_parameters):
        super().__init__()
        self.image_parameters = image_parameters
        self.grid_parameters = grid_parameters
        self.kernel = ConvolutionKernel(image_parameters, grid_parameters)
        pixels = image_parameters.pixels
        shape = (len(image_parameters.polarizations), pixels, pixels)
        self.values = np.empty(shape, image_parameters.complex_dtype)


class GridderHost(GridDegridHost):
    def __init__(self, image_parameters, grid_parameters):
        super().__init__(image_parameters, grid_parameters)
        self.weights_grid = np.empty(self.values.shape, np.float32)

    def clear(self):
        self.values.fill(0)

    def __call__(self):
        """Add visibilities to the grid, with convolutional gridding.

        Parameters
        ----------
        """
        _grid(self.kernel.data, self.values, self.weights_grid,
              self.uv, self.sub_uv, self.w_plane, self.vis,
              np.empty((self.vis.shape[1],), self.values.dtype))


@numba.jit(nopython=True)
def _degrid(kernel, values, uv, sub_uv, w_plane, weights, vis, sample):
    ksize = kernel.shape[2]
    uv_bias = (ksize - 1) // 2 - values.shape[2] // 2
    for row in range(uv.shape[0]):
        u0 = uv[row, 0] - uv_bias
        v0 = uv[row, 1] - uv_bias
        sub_u, sub_v = sub_uv[row]
        for i in range(vis.shape[1]):
            sample[i] = 0
        for j in range(ksize):
            for k in range(ksize):
                weight = kernel[w_plane[row], sub_v, j] * kernel[w_plane[row], sub_u, k]
                for pol in range(values.shape[0]):
                    sample[pol] += weight * values[pol, v0 + j, u0 + k]
        for i in range(vis.shape[1]):
            vis[row, i] -= weights[row, i] * sample[i]


class DegridderHost(GridDegridHost):
    def __init__(self, image_parameters, grid_parameters):
        super().__init__(image_parameters, grid_parameters)
        self.weights = None

    @VisOperationHost.num_vis.setter
    def num_vis(self, value):
        VisOperationHost.num_vis.fset(self, value)
        self.weights = None

    def set_weights(self, weights):
        """Set statistical weights"""
        if len(weights) != self.num_vis:
            raise ValueError('Lengths do not match')
        self.weights = weights

    def __call__(self):
        """Compute visibilities from the grid. See :meth:`GridderHost.grid`
        for details of the parameters.
        """
        _degrid(self.kernel.data, self.values,
                self.uv, self.sub_uv, self.w_plane, self.weights, self.vis,
                np.empty((self.vis.shape[1],), self.values.dtype))



"""Tests for :mod:`katsdpimager.image`"""

import math

import numpy as np
from katsdpsigproc.test.test_accel import device_test

from .. import image
from .utils import RandomState


class TestLayerToImage:
    @device_test
    def test(self, context, command_queue):
        slices = 3
        size = 102
        shape = (slices, size, size)
        lm_scale = 0.1 / size
        lm_bias = -lm_scale * size / 3   # Off-centre, to check that it's working
        w = 12.3
        template = image.LayerToImageTemplate(context, np.float32)
        fn = template.instantiate(command_queue, shape, lm_scale, lm_bias)
        fn.set_w(w)
        fn.set_polarization(1)
        fn.ensure_all_bound()
        # Create random input data
        rs = RandomState(1)
        src = rs.complex_uniform(10.0, 100.0, shape[1:]).astype(np.complex64)
        kernel1d = rs.uniform(1.0, 2.0, size).astype(np.float32)
        fn.buffer('kernel1d').set(command_queue, kernel1d)
        fn.buffer('layer').set(command_queue, src)
        fn.buffer('image').zero(command_queue)
        # Compute expected value
        lm = np.arange(size) * lm_scale + lm_bias
        lm2 = lm * lm
        n = np.sqrt(1 - lm2[np.newaxis, :, np.newaxis] - lm2[np.newaxis, np.newaxis, :])
        w_correction = np.exp(2j * math.pi * w * (n - 1))
        corrected = np.fft.fftshift(src) * w_correction
        expected = np.zeros(shape, np.float32)
        expected[1] = corrected.real * n / np.outer(kernel1d, kernel1d)[np.newaxis, ...]
        # Check it
        fn()
        actual = fn.buffer('image').get(command_queue)
        np.testing.assert_allclose(expected, actual, 1e-4)


class TestImageToLayer:
    @device_test
    def test(self, context, command_queue):
        # Since LayerToImage is tested from first principles, we just check
        # that ImageToLayer correctly inverts it
        slices = 3
        size = 102
        shape = (slices, size, size)
        lm_scale = 0.1 / size
        lm_bias = -lm_scale * size / 3   # Off-centre, to check that it's working
        w = 12.3
        image_to_layer_template = image.ImageToLayerTemplate(context, np.float32)
        image_to_layer = image_to_layer_template.instantiate(
            command_queue, shape, lm_scale, lm_bias)
        image_to_layer.ensure_all_bound()
        image_to_layer.set_w(w)
        layer_to_image_template = image.LayerToImageTemplate(context, np.float32)
        layer_to_image = layer_to_image_template.instantiate(
            command_queue, shape, lm_scale, lm_bias)
        layer_to_image.set_w(w)
        layer_to_image.bind(
            image=image_to_layer.buffer('image'),
            layer=image_to_layer.buffer('layer'),
            kernel1d=image_to_layer.buffer('kernel1d'))
        # Create random input data
        rs = np.random.RandomState(1)
        expected = rs.uniform(10.0, 100.0, shape).astype(np.float32)
        kernel1d = rs.uniform(1.0, 2.0, size).astype(np.float32)
        kernel = np.outer(kernel1d, kernel1d)[np.newaxis, ...]
        # Run test function
        image_to_layer.buffer('image').set(command_queue, expected * kernel)
        image_to_layer.buffer('kernel1d').set(command_queue, kernel1d)
        image_to_layer.set_polarization(1)
        image_to_layer()
        # Wipe out the image, to make sure it gets written properly
        image_to_layer.buffer('image').zero(command_queue)
        # Convert back again
        layer_to_image.set_polarization(1)
        layer_to_image()
        actual = layer_to_image.buffer('image').get(command_queue) * kernel
        np.testing.assert_allclose(expected[1], actual[1], 1e-4)


class TestScale:
    @device_test
    def test(self, context, command_queue):
        shape = (4, 123, 234)
        dtype = np.float32
        rs = np.random.RandomState(1)
        template = image.ScaleTemplate(context, dtype, shape[0])
        fn = template.instantiate(command_queue, shape)
        fn.ensure_all_bound()
        src = rs.uniform(size=shape).astype(np.float32)
        scale_factor = np.array([1.2, 2.3, 3.4, -4.5], dtype)
        data = fn.buffer('data')
        data.set(command_queue, src)
        fn.set_scale_factor(scale_factor)
        fn()
        actual = data.get(command_queue)
        expected = src * scale_factor[:, np.newaxis, np.newaxis]
        np.testing.assert_allclose(expected, actual)



"""Tests for :py:mod:`katsdpimager.beam`"""

import numpy as np
import scipy.signal
from astropy.modeling import models
from katsdpsigproc.test.test_accel import device_test, cuda_test

from katsdpimager import beam


def convolve_beam_reference(model, beam):
    """Reference implementation of
    :py:func:`~katsdpimager.beam.convolve_beam`, for comparison.

    It will not produce exactly the same result, because it uses a sampled
    form of the beam, and does not wrap around.
    """
    assert model.shape[1] % 2 == 0
    assert model.shape[2] % 2 == 0
    hh = model.shape[1] // 2
    hw = model.shape[2] // 2
    m = np.arange(-hh + 1, hh)
    l = np.arange(-hw + 1, hw)
    beam_pixels = beam.model(*np.meshgrid(m, l, indexing='ij'))
    out = np.empty_like(model)
    for pol in range(model.shape[0]):
        out[pol, ...] = scipy.signal.fftconvolve(model[pol], beam_pixels, 'same')
    return out


class TestConvolveBeam:
    """Test both host and device beam convolution functions"""
    def setup(self):
        self.beam = beam.Beam(models.Gaussian2D(amplitude=3.5, x_mean=0.0, y_mean=0.0,
                                                x_stddev=2.0, y_stddev=5.0, theta=1))
        self.model = np.zeros((4, 128, 128), np.float32)
        self.model[0, 32, 80] = 1.0
        self.model[0, 100, 40] = 2.0
        self.model[1, 50, 60] = 3.0
        self.model[2, 64, 64] = 4.0
        self.model[2, 80, 64] = 3.0
        self.expected = convolve_beam_reference(self.model, self.beam)

    def test_host(self):
        actual = beam.convolve_beam(self.model, self.beam)
        np.testing.assert_allclose(self.expected, actual, rtol=1e-5, atol=1e-5)

    @device_test
    @cuda_test
    def test_device(self, context, command_queue):
        template = beam.ConvolveBeamTemplate(command_queue, self.model.shape[1:], self.model.dtype)
        fn = template.instantiate()
        fn.ensure_all_bound()
        for pol in range(self.model.shape[0]):
            fn.buffer('image').set(command_queue, self.model[pol])
            fn.beam = self.beam
            fn()
            actual = fn.buffer('image').get(command_queue)
            np.testing.assert_allclose(self.expected[pol], actual, rtol=1e-5, atol=1e-5)



"""Tests for :mod:`katsdpimager.weight`."""

import numpy as np
import katsdpsigproc.accel as accel
from katsdpsigproc.test.test_accel import device_test

from katsdpimager import weight


class TestGridWeights:
    def setup(self):
        self.grid_shape = (4, 100, 200)
        self.uv = np.array([
            [-10, 5, 0, 0],
            [23, 17, 0, 0],
            [-10, 5, 0, 0],
            [-10, 5, 0, 0],
            [-10, 6, 0, 0],
            [-11, 5, 0, 0]], np.int16)
        self.weights = np.array([
            [1.0, 10.0, 100.0, 1000.0],
            [2.0, 20.0, 200.0, 2000.0],
            [4.0, 40.0, 400.0, 4000.0],
            [8.0, 80.0, 800.0, 8000.0],
            [16.0, 160.0, 1600.0, 16000.0],
            [32.0, 320.0, 3200.0, 32000.0]], np.float32)

    def _set_partial(self, command_queue, buffer, data):
        """Load a device array from a smaller host array. The rest of the
        array is filled with random values to ensure that the algorithm
        doesn't depend on them being zero.
        """
        rs = np.random.RandomState(1)
        host = buffer.empty_like()
        host[:] = rs.uniform(size=host.shape)
        sub_rect = tuple(np.s_[0 : x] for x in data.shape)
        host[sub_rect] = data
        buffer.set(command_queue, host)

    @device_test
    def test(self, context, command_queue):
        template = weight.GridWeightsTemplate(context, 4)
        fn = template.instantiate(command_queue, self.grid_shape, 1000)
        fn.ensure_all_bound()
        self._set_partial(command_queue, fn.buffer('uv'), self.uv)
        self._set_partial(command_queue, fn.buffer('weights'), self.weights)
        fn.buffer('grid').zero(command_queue)
        fn.num_vis = len(self.uv)
        fn()
        actual = fn.buffer('grid').get(command_queue)
        expected = np.zeros(self.grid_shape, np.float32)
        for i in range(4):
            expected[i, 55, 90] = 13 * 10**i
            expected[i, 67, 123] = 2 * 10**i
            expected[i, 56, 90] = 16 * 10**i
            expected[i, 55, 89] = 32 * 10**i
        np.testing.assert_equal(expected, actual)


class TestDensityWeights:
    def setup(self):
        rs = np.random.RandomState(1)
        self.grid_shape = (4, 50, 107)
        self.data = np.zeros(self.grid_shape, np.float32)
        self.expected = np.zeros(self.grid_shape, np.float32)
        n_indices = 100
        flat_indices = rs.choice(self.data.size, n_indices, replace=False)
        sum_w = 0.0
        sum_dw = 0.0
        sum_d2w = 0.0
        for index in flat_indices:
            w = rs.uniform(low=0.1, high=2.0)
            d = 1.0 / (2.5 * w + 1.75)
            self.data.flat[index] = w
            self.expected.flat[index] = d
            if index < self.data[0].size:
                sum_w += w
                sum_dw += d * w
                sum_d2w += d**2 * w
        self.normalized_rms = np.sqrt(sum_d2w * sum_w) / sum_dw

    @device_test
    def test(self, context, command_queue):
        template = weight.DensityWeightsTemplate(context, 4)
        fn = template.instantiate(command_queue, self.grid_shape)
        fn.ensure_all_bound()
        fn.a = 2.5
        fn.b = 1.75
        # The 'grid' is set in a roundabout way to fill the padding with
        # non-zero values
        tmp_data = fn.buffer('grid').empty_like()
        accel.HostArray.padded_view(tmp_data).fill(3)
        tmp_data[:] = self.data
        fn.buffer('grid').set(command_queue, tmp_data)
        normalized_rms = fn()
        actual = fn.buffer('grid').get(command_queue)
        np.testing.assert_allclose(self.expected, actual, 1e-5, 1e-5)
        np.testing.assert_allclose(self.normalized_rms, normalized_rms, 1e-6)


class TestMeanWeight:
    def setup(self):
        rs = np.random.RandomState(1)
        self.grid_shape = (4, 50, 107)
        self.data = rs.uniform(size=self.grid_shape).astype(np.float32)

    @device_test
    def test(self, context, command_queue):
        template = weight.MeanWeightTemplate(context)
        fn = template.instantiate(command_queue, self.grid_shape)
        fn.ensure_all_bound()
        fn.buffer('grid').set(command_queue, self.data)
        actual = fn()
        pol0 = self.data[0]
        expected = np.sum(pol0 * pol0) / np.sum(pol0)
        np.testing.assert_allclose(expected, actual, rtol=1e-5)



"""Wrapper around numba that either uses numba, or provides a no-op jit
function."""

try:
    from numba import jit
    have_numba = True
except ImportError:
    def jit(*args, **kwargs):
        return lambda x: x
    have_numba = False



"""Utilities to wrap skcuda.fft for imaging purposes

.. include:: macros.rst
"""

import numpy as np
import pkg_resources
import skcuda.fft
from katsdpsigproc import accel

import katsdpimager.types

#: Forward FFT
FFT_FORWARD = 0
#: Inverse FFT, without normalization factor of 1/N
FFT_INVERSE = 1


class _Gpudata:
    """Adapter to allow skcuda.fft to work with managed memory
    allocations. Add it as a `gpudata` member on an arbitrary object, to allow
    that object to be passed instead of a :py:class`pycuda.gpuarray.GPUArray`.
    """
    def __init__(self, array):
        # .buffer gives the ndarray created by PyCUDA, and .base the ManagedAllocation
        self._allocation = array.buffer.base

    def __int__(self):
        return self._allocation.get_device_pointer()

    def __eq__(self, other):
        return int(self) == int(other)

    def __ne__(self, other):
        return int(self) != int(other)


class _GpudataWrapper:
    """Forwarding wrapper around a :py:class:`katsdpsigproc.accel.SVMArray` or
    :py:class:`katsdpsigproc.accel.DeviceArray` that allows it to be passed
    to skcuda.fft.
    """
    def __init__(self, wrapped):
        self._wrapped = wrapped
        try:
            # Handle DeviceArray case
            self.gpudata = wrapped.buffer.gpudata
        except AttributeError:
            # SVMArray case
            self.gpudata = _Gpudata(wrapped)

    def __getattr__(self, attr):
        return getattr(self._wrapped, attr)


class FftshiftTemplate:
    """Operation template for the equivalent of :py:meth:`np.fft.fftshift` on
    the device, in-place. The last two dimensions are shifted, and these
    dimensions must have even size. Because the size is even, this operation
    is its own inverse.

    Parameters
    ----------
    context : |Context|
        Context for which kernels will be compiled
    dtype : numpy dtype
        Data type being stored
    tuning : dict, optional
        Tuning parameters (currently unused)
    """
    def __init__(self, context, dtype, tuning=None):
        self.context = context
        self.dtype = np.dtype(dtype)
        # TODO: autotune
        self.wgsx = 16
        self.wgsy = 8
        self.program = accel.build(
            context, "imager_kernels/fftshift.mako",
            {
                'ctype': katsdpimager.types.dtype_to_ctype(dtype),
                'wgsx': self.wgsx,
                'wgsy': self.wgsy
            },
            extra_dirs=[pkg_resources.resource_filename(__name__, '')])

    def instantiate(self, *args, **kwargs):
        return Fftshift(self, *args, **kwargs)


class Fftshift(accel.Operation):
    """Instantiation of :py:class:`FftshiftTemplate`.

    .. rubric:: Slots

    **data**
        Input and output array, transformed in-place

    Parameters
    ----------
    template : :class:`FftshiftTemplate`
        Operation template
    command_queue : |CommandQueue|
        Command queue for the operation
    shape : tuple of int
        Shape of the data.
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots

    Raises
    ------
    ValueError
        if the first two dimensions in `shape` are not even
    """
    def __init__(self, template, command_queue, shape, allocator=None):
        super().__init__(command_queue, allocator)
        self.template = template
        self.kernel = template.program.get_kernel('fftshift')
        if shape[-1] % 2 != 0 or shape[-2] % 2 != 0:
            raise ValueError('First two dimensions of shape must be even')
        self.slots['data'] = accel.IOSlot(shape, template.dtype)

    def _run(self):
        data = self.buffer('data')
        items_x = data.shape[-1] // 2
        items_y = data.shape[-2] // 2
        if len(data.shape) == 2:
            items_z = 1
        else:
            items_z = data.shape[0] * int(np.product(data.padded_shape[1:-2]))
        row_stride = data.padded_shape[-1]
        slice_stride = data.padded_shape[-2] * data.padded_shape[-1]
        self.command_queue.enqueue_kernel(
            self.kernel,
            [
                data.buffer,
                np.int32(row_stride),
                np.int32(slice_stride),
                np.int32(items_x), np.int32(items_y),
                np.int32(items_y * row_stride)
            ],
            global_size=(accel.roundup(items_x, self.template.wgsx),
                         accel.roundup(items_y, self.template.wgsy),
                         items_z),
            local_size=(self.template.wgsx, self.template.wgsy, 1)
        )


class FftTemplate:
    r"""Operation template for a forward or reverse FFT. The transformation is
    done over the last N dimensions, with the remaining dimensions for batching
    multiple arrays to be transformed. Dimensions before the first N must have
    consistent padding between the source and destination, and it is
    recommended to have no padding at all since the padding arrays are also
    transformed.

    This template bakes in more information than most (command queue and data
    shapes), which is due to constraints in CUFFT.

    The template can specify real->complex, complex->real, or
    complex->complex. In the last case, the same template can be used to
    instantiate forward or inverse transforms. Otherwise, real->complex
    transforms must be forward, and complex->real transforms must be inverse.

    For real<->complex transforms, the final dimension of the padded shape
    need only be :math:`\lfloor\frac{L}{2}\rfloor + 1`, where :math:`L` is the
    last element of `shape`.

    Parameters
    ----------
    command_queue : :class:`katsdpsigproc.cuda.CommandQueue`
        Command queue for the operation
    N : int
        Number of dimensions for the transform
    shape : tuple
        Shape of the input (N or more dimensions)
    dtype_src : {`np.float32, `np.float64`, `np.complex64`, `np.complex128`}
        Data type for input
    dtype_dest : {`np.float32, `np.float64`, `np.complex64`, `np.complex128`}
        Data type for output
    padded_shape_src : tuple
        Padded shape of the input
    padded_shape_dest : tuple
        Padded shape of the output
    tuning : dict, optional
        Tuning parameters (currently unused)
    """
    def __init__(self, command_queue, N, shape, dtype_src, dtype_dest,
                 padded_shape_src, padded_shape_dest, tuning=None):
        if padded_shape_src[:-N] != padded_shape_dest[:-N]:
            raise ValueError('Source and destination padding does not match on batch dimensions')
        self.command_queue = command_queue
        self.shape = shape
        self.dtype_src = np.dtype(dtype_src)
        self.dtype_dest = np.dtype(dtype_dest)
        self.padded_shape_src = padded_shape_src
        self.padded_shape_dest = padded_shape_dest
        # CUDA 7.0 CUFFT has a bug where kernels are run in the default stream
        # instead of the requested one, if dimensions are up to 1920. There is
        # a patch, but there is no query to detect whether it has been
        # applied.
        self.needs_synchronize_workaround = any(x <= 1920 for x in shape[:N])
        batches = int(np.product(padded_shape_src[:-N]))
        with command_queue.context:
            self.plan = skcuda.fft.Plan(
                shape[-N:], dtype_src, dtype_dest, batches,
                stream=command_queue._pycuda_stream,
                inembed=np.array(padded_shape_src[-N:], np.int32),
                istride=1,
                idist=int(np.product(padded_shape_src[-N:])),
                onembed=np.array(padded_shape_dest[-N:], np.int32),
                ostride=1,
                odist=int(np.product(padded_shape_dest[-N:])))

    def instantiate(self, *args, **kwargs):
        return Fft(self, *args, **kwargs)


class Fft(accel.Operation):
    """Forward or inverse Fourier transformation.

    .. rubric:: Slots

    **src**
        Input data
    **dest**
        Output data

    Parameters
    ----------
    template : :class:`FftTemplate`
        Operation template
    mode : {:data:`FFT_FORWARD`, :data:`FFT_INVERSE`}
        FFT direction
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots
    """
    def __init__(self, template, mode, allocator=None):
        super().__init__(template.command_queue, allocator)
        self.template = template
        src_shape = list(template.shape)
        dest_shape = list(template.shape)
        if template.dtype_src.kind != 'c':
            if mode != FFT_FORWARD:
                raise ValueError('R2C transform must use FFT_FORWARD')
            dest_shape[-1] = template.shape[-1] // 2 + 1
        if template.dtype_dest.kind != 'c':
            if mode != FFT_INVERSE:
                raise ValueError('C2R transform must use FFT_INVERSE')
            src_shape[-1] = template.shape[-1] // 2 + 1
        src_dims = [accel.Dimension(d[0], min_padded_size=d[1], exact=True)
                    for d in zip(src_shape, template.padded_shape_src)]
        dest_dims = [accel.Dimension(d[0], min_padded_size=d[1], exact=True)
                     for d in zip(dest_shape, template.padded_shape_dest)]
        self.slots['src'] = accel.IOSlot(src_dims, template.dtype_src)
        self.slots['dest'] = accel.IOSlot(dest_dims, template.dtype_dest)
        self.mode = mode

    def _run(self):
        src_buffer = self.buffer('src')
        dest_buffer = self.buffer('dest')
        context = self.template.command_queue.context
        with context:
            if self.mode == FFT_FORWARD:
                skcuda.fft.fft(_GpudataWrapper(src_buffer), _GpudataWrapper(dest_buffer),
                               self.template.plan)
            else:
                skcuda.fft.ifft(_GpudataWrapper(src_buffer), _GpudataWrapper(dest_buffer),
                                self.template.plan)
            if self.template.needs_synchronize_workaround:
                context._pycuda_context.synchronize()



"""Assorted equations for computing appropriate imaging parameters. The
functions take physical quantities as Astropy quantities, rather than
specifying any specific units.

Most formulae are taken from SKA-TEL-SDP-0000003.
"""

import math

from astropy import units
import numpy as np

import katsdpimager.types
from . import clean, weight


def is_smooth(x):
    """Whether x is a good candidate for FFT. We heuristically require
    it to be a multiple of 8 and a product of powers of 2, 3, 5 and 7."""
    if x % 8 != 0:
        return False
    for d in [2, 3, 5, 7]:
        while x % d == 0:
            x = x // d
    return x == 1


class ArrayParameters:
    """Physical attributes of an interferometric array."""
    def __init__(self, antenna_diameter, longest_baseline):
        assert antenna_diameter.unit.physical_type == 'length'
        assert longest_baseline.unit.physical_type == 'length'
        self.antenna_diameter = antenna_diameter
        self.longest_baseline = longest_baseline


class ImageParameters:
    """Physical properties associated with an image. At present, only
    single-frequency images are supported.

    Parameters
    ----------
    q_fov : float
        Scale factor for field of view. 1.0 specifies the first null of the
        primary beam (computed just from an Airy disk, not a measured beam
        model).
    image_oversample : float
        Number of pixels per beam
    frequency : Quantity
        Representative frequency, for converting UVW between metres and
        wavelengths. It may also be specified as a wavelength.
    array : :class:`ArrayParameters`
        Properties of the array. It is not needed if both `pixel_size` and
        `pixels` are specified.
    polarizations : list
        List of polarizations that will appear in the image
    dtype : {np.float32, np.complex64}
        Floating-point type for image and grid
    pixel_size : Quantity or float, optional
        Angular size of a single pixel, or dimensionless to specify l or m
        size directly. If specified, `image_oversample` is ignored.
    pixels : int, optional
        Number of pixels in the image. If specified, `q_fov` is ignored.
    """
    def __init__(self, q_fov, image_oversample, frequency, array, polarizations,
                 dtype, pixel_size=None, pixels=None):
        self.wavelength = frequency.to(units.m, equivalencies=units.spectral())
        # Compute pixel size
        if pixel_size is None:
            if image_oversample < 3.0:
                raise ValueError('image_oversample is too small '
                                 'to capture all visibilities in the UV plane')
            # Size of the UV image plane
            uv_size = (2.0 / 3.0 * image_oversample) * array.longest_baseline
            self.pixel_size = (self.wavelength / uv_size).decompose()
        else:
            # Ensure pixel_size is a Quantity and not just a float
            pixel_size = units.Quantity(pixel_size)
            if pixel_size.unit.physical_type == 'angle':
                pixel_size = np.sin(pixel_size)
            self.pixel_size = pixel_size
        # Compute number of pixels
        if pixels is None:
            # These are just a preliminary cell and pixel size, to compute pixels
            cell_size = array.antenna_diameter * (math.pi / (7.6634 * q_fov))
            image_size = self.wavelength / cell_size
            # Allow image to be slightly smaller if it makes the Fourier transform easier
            pixels = int(0.98 * image_size / self.pixel_size)
            while not is_smooth(pixels):
                pixels += 1
        else:
            if not is_smooth(pixels):
                recommended = pixels
                while not is_smooth(recommended):
                    recommended += 1
                raise ValueError("Image size {} not supported - try {}".format(pixels, recommended))
        assert pixels % 2 == 0
        self.real_dtype = np.dtype(dtype)
        self.complex_dtype = katsdpimager.types.real_to_complex(dtype)
        self.pixels = pixels
        self.image_size = self.pixel_size * pixels
        self.cell_size = self.wavelength / self.image_size
        self.polarizations = polarizations

    def __str__(self):
        from . import polarization
        return """\
Pixel size: {:.3f}
Pixels: {}
FOV: {:.3f}
Cell size: {:.3f}
Wavelength: {:.3f}
Polarizations: {}
Precision: {} bit
""".format(np.arcsin(self.pixel_size).to(units.arcsec),
           self.pixels,
           np.arcsin(self.pixel_size * self.pixels).to(units.deg),
           self.cell_size, self.wavelength,
           ','.join([polarization.STOKES_NAMES[i] for i in self.polarizations]),
           32 if self.real_dtype == np.float32 else 64)


def w_kernel_width(image_parameters, w, eps_w, antialias_width=0):
    """Determine the width (in UV cells) for a W kernel. This is Eq 9 of
    SKA-TEL-SDP-0000003.

    Parameters
    ----------
    image_parameters : :class:`ImageParameters`
        Image parameters, from which wavelength and image size are used
    w : Quantity
        W value for the kernel, as a distance
    eps_w : float
        Fraction of peak at which to truncate the kernel
    antialias_width : float, optional
        If provided, the return value is for a combined W and antialias
        kernel, where the sizes of the individual kernels are combined
        in quadrature.
    """
    fov = image_parameters.image_size
    wl = float(w / image_parameters.wavelength)
    # Squared size of the w part
    wk2 = 4 * fov**2 * (
        (wl * image_parameters.image_size / 2)**2
        + wl**1.5 * fov / (2 * math.pi * eps_w))
    return np.sqrt(wk2 + antialias_width**2)


def w_slices(image_parameters, max_w, eps_w, kernel_width, antialias_width=0):
    lo = 0
    hi = 1
    # Each slice is corrected to its center, so maximum W deviation from the
    # slice center is only half the slice thickness.
    max_w = max_w * 0.5
    # Find a number of slices that is definitely big enough. The first slice is
    # only half-width, to allow the (possibly numerous) visibilities with small
    # W to have better accuracy.

    def measure(slices):
        return w_kernel_width(image_parameters, max_w / (slices - 0.5), eps_w, antialias_width)

    while measure(hi) > kernel_width:
        hi *= 2
    # Binary search
    while hi - lo > 1:
        mid = (lo + hi) // 2
        if measure(mid) < kernel_width:
            hi = mid
        else:
            lo = mid
    return hi


class WeightParameters:
    """Parameters affecting imaging weight calculations.

    Parameters
    ----------
    weight_type : :class:`weight.WeightType`
        Image weighting scheme
    robustness : float, optional
        Robustness parameter for robust weighting
    """   # noqa: E501
    def __init__(self, weight_type, robustness=0.0):
        self.weight_type = weight_type
        self.robustness = robustness

    def __str__(self):
        if self.weight_type == weight.WeightType.ROBUST:
            ans = 'robust ({:.3f})'.format(self.robustness)
        else:
            ans = self.weight_type.name.lower()
        return 'Image weights: ' + ans


class GridParameters:
    """Parameters affecting gridding algorithm.

    Parameters
    ----------
    antialias_width : float
        Support of the antialiasing kernel
    oversample : int
        Number of UV sub-cells per cell, for sampling kernels
    image_oversample : int
        Oversampling in image plane during kernel generation
    w_slices : int
        Number of slices for w-stacking
    w_planes : int
        Number of samples to take in w within each slice
    max_w : Quantity
        Maximum absolute w value, as a distance quantity
    kernel_width : int
        Number of UV cells corresponding to the combined W+antialias kernel.
    degrid : bool, optional
        If true, use degridding, otherwise use direct prediction.
    """
    def __init__(self, antialias_width, oversample, image_oversample,
                 w_slices, w_planes, max_w, kernel_width, degrid=False):
        if max_w.unit.physical_type != 'length':
            raise TypeError('max W must be specified as a length')
        self.antialias_width = antialias_width
        self.oversample = oversample
        self.image_oversample = image_oversample
        self.w_slices = w_slices
        self.w_planes = w_planes
        self.max_w = max_w
        self.kernel_width = kernel_width
        self.degrid = degrid

    def __str__(self):
        prediction = 'degridding' if self.degrid else 'direct'
        return """\
Grid oversampling: {self.oversample}
Image oversample: {self.image_oversample}
W slices: {self.w_slices}
W planes per slice: {self.w_planes}
Maximum W: {self.max_w:.3f}
Antialiasing support: {self.antialias_width} cells
Kernel support: {self.kernel_width} cells
Prediction: {prediction}""".format(self=self, prediction=prediction)


class CleanParameters:
    def __init__(self, minor, loop_gain, major_gain, threshold, mode,
                 psf_cutoff, psf_limit, border):
        self.minor = minor
        self.loop_gain = loop_gain
        self.major_gain = major_gain
        self.threshold = threshold
        self.mode = mode
        self.psf_cutoff = psf_cutoff
        if self.psf_cutoff >= 1.0:
            raise ValueError('PSF cutoff must be less than 1')
        self.psf_limit = psf_limit
        self.border = border

    def __str__(self):
        return """\
Loop gain: {self.loop_gain}
Major cycle gain: {self.major_gain}
Threshold: {self.threshold} sigma
Max minor cycles: {self.minor}
PSF cutoff: {self.psf_cutoff}
PSF limit: {self.psf_limit} pixels
Peak function: {mode}
Border: {self.border} pixels""".format(
            self=self, mode='I' if self.mode == clean.CLEAN_I else 'I^2+Q^2+U^2+V^2')



#!/usr/bin/env python

"""
Creates an HTML page showing the images written by the Jenkins script.
"""

import argparse
import os
import os.path
import shutil
import sys
import subprocess
import logging
import timeit
import io
import glob
from contextlib import closing

from mako.template import Template
import matplotlib
matplotlib.use('Agg')   # noqa: E402
import aplpy
from astropy.io import fits
from astropy.wcs import WCS


MODE_CLEAN = 'Clean'
MODE_DIRTY = 'Dirty'
MODES = [MODE_CLEAN, MODE_DIRTY]


class BuildInfo:
    """Runs a command and stores the output and exit code. The output is also
    echoed to stdout.
    """
    def __init__(self, cmds):
        self.cmds = cmds
        start = timeit.default_timer()
        output = io.BytesIO()
        self.returncode = 0
        for cmd in cmds:
            sys.stdout.flush()
            child = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            # The child.stdout seems to do full reads, even though it is supposedly unbuffered
            stdout = io.open(child.stdout.fileno(), mode='rb', buffering=0, closefd=False)
            sys_stdout = io.open(sys.stdout.fileno(), mode='wb', buffering=0, closefd=False)
            while True:
                data = stdout.read(256)
                if not data:
                    break
                output.write(data)
                sys_stdout.write(data)
                sys_stdout.flush()
            self.returncode = child.wait()
            if self.returncode != 0:
                logging.warning("Process failed with exit code %s", self.returncode)
                break
        self.output = output.getvalue().decode('utf-8')
        end = timeit.default_timer()
        self.elapsed = end - start
        logging.info("Elapsed time %.3f", self.elapsed)


class Image:
    def __init__(self, name, filebase,
                 clean_fits_pattern, dirty_fits_pattern, cmd_patterns,
                 clean_globs=[]):
        self.name = name
        self.filebase = filebase
        self.fits_patterns = {MODE_CLEAN: clean_fits_pattern, MODE_DIRTY: dirty_fits_pattern}
        self.cmd_patterns = cmd_patterns
        self.clean_globs = list(clean_globs)

    def fits_filename(self, mode, stokes, channel, rel_channel):
        return self.fits_patterns[mode].format(stokes=stokes, channel=channel,
                                               rel_channel=rel_channel)

    def svg_filename_thumb(self, mode, stokes, channel):
        return '{}-{}-{}-{}-thumb.svg'.format(self.filebase, mode, stokes, channel)

    def svg_filename_full(self, mode, stokes, channel):
        return '{}-{}-{}-{}.svg'.format(self.filebase, mode, stokes, channel)

    def build_info_filename(self):
        return '{}-log.html'.format(self.filebase)

    def build(self, ms, output_dir, modes, stokes, channels):
        cmds = [[
                Template(x).render_unicode(
                    ms=ms, output_dir=output_dir, stokes=stokes, channels=channels)
                for x in pattern
                ] for pattern in self.cmd_patterns]
        logging.info("Running %s...", cmds)
        build_info = BuildInfo(cmds)
        # Clean up unwanted files, making sure not to delete any files we
        # actually wanted.
        keep = set(os.path.join(output_dir,
                                self.fits_filename(mode, s, channel, channel - channels[0]))
                   for mode in modes
                   for s in stokes
                   for channel in channels)
        for clean_glob in self.clean_globs:
            for filename in glob.iglob(os.path.join(output_dir, clean_glob)):
                if filename not in keep:
                    logging.info('Removing intermediate file %s', filename)
                    if os.path.isdir(filename):
                        shutil.rmtree(filename)
                    else:
                        os.remove(filename)
        return build_info

    def _render_common(self, figure, colorscale_args={}):
        figure.show_colorscale(vmin=-0.01, vmax=2.0, vmid=-0.015, stretch='log', **colorscale_args)
        figure.add_colorbar()
        figure.add_grid()

    def _render_thumb(self, output_dir, input_filename, slices, mode, stokes, channel):
        figure = aplpy.FITSFigure(input_filename, slices=slices, figsize=(4, 4))
        self._render_common(figure)
        figure.colorbar.set_font(size='x-small')
        figure.tick_labels.set_font(size='xx-small')
        figure.save(os.path.join(output_dir, self.svg_filename_thumb(mode, stokes, channel)))
        figure.close()

    def _render_full(self, output_dir, input_filename, slices, mode, stokes, channel):
        figure = aplpy.FITSFigure(input_filename, slices=slices, figsize=(19, 18))
        self._render_common(figure, colorscale_args=dict(interpolation=None))
        figure.colorbar.set_font(size='small')
        figure.tick_labels.set_font(size='small')
        figure.save(os.path.join(output_dir, self.svg_filename_full(mode, stokes, channel)))
        figure.close()

    def render(self, output_dir, mode, stokes, channel, rel_channel):
        # Check whether the Stokes parameters and/or channel are all bundled
        # into one image, and if so, slice it.
        filename = os.path.join(output_dir, self.fits_filename(mode, stokes, channel, rel_channel))
        with closing(fits.open(filename)) as hdulist:
            wcs = WCS(hdulist[0])
            wcsaxes = wcs.naxis
            # Reverse shape to put back into same order as axis_types
            shape = list(reversed(hdulist[0].shape))
            if wcsaxes < 2:
                raise ValueError('At least two dimensions expected')
            axis_types = wcs.get_axis_types()
            if axis_types[0]['coordinate_type'] != 'celestial' or axis_types[0]['number'] != 0:
                raise ValueError('First axis is not longitudinal')
            if axis_types[1]['coordinate_type'] != 'celestial' or axis_types[1]['number'] != 1:
                raise ValueError('Second axis is not latitudinal')
            slices = [0] * (len(shape) - 2)
            for i in range(2, len(shape)):
                ctype = axis_types[i]['coordinate_type']
                if ctype == 'stokes':
                    # TODO: should use the WCS transformation
                    slices[i - 2] = 'IQUV'.find(stokes)
                elif ctype == 'spectral' and shape[i] > 1:
                    slices[i - 2] = rel_channel
        self._render_thumb(output_dir, filename, slices, mode, stokes, channel)
        self._render_full(output_dir, filename, slices, mode, stokes, channel)


def write_index(args, images, build_info, modes):
    template_filename = os.path.join(os.path.dirname(__file__), 'report.html.mako')
    template = Template(filename=template_filename)
    try:
        commit = subprocess.check_output(['git', 'rev-parse', 'HEAD'])
    except subprocess.CalledProcessError:
        commit = 'Unknown'
    with io.open(os.path.join(args.output_dir, 'index.html'), 'w', encoding='utf-8') as f:
        f.write(template.render_unicode(
            revision=commit,
            modes=modes,
            stokes=args.stokes,
            channels=list(range(args.start_channel, args.stop_channel)),
            images=images,
            build_info=build_info))


def write_build_log(args, image, build_info, modes):
    template_filename = os.path.join(os.path.dirname(__file__), 'build_log.html.mako')
    template = Template(filename=template_filename)
    path = os.path.join(args.output_dir, image.build_info_filename())
    with io.open(path, 'w', encoding='utf-8') as f:
        f.write(template.render_unicode(
            image=image, build_info=build_info, modes=modes, stokes=args.stokes,
            channels=list(range(args.start_channel, args.stop_channel))))


def run(args, images, modes):
    try:
        os.makedirs(args.output_dir)
    except OSError:
        if not os.path.isdir(args.output_dir):
            raise
    build_info = {}
    channels = list(range(args.start_channel, args.stop_channel))
    for image in images:
        if not args.skip_build:
            build_info[image] = image.build(args.ms, args.output_dir, modes, args.stokes, channels)
        else:
            build_info[image] = BuildInfo([])

    write_index(args, images, build_info, modes)
    for image in images:
        write_build_log(args, image, build_info[image], modes)

    for image in images:
        if build_info[image].returncode == 0:
            for mode in modes:
                for stokes in args.stokes:
                    for channel in channels:
                        image.render(args.output_dir, mode, stokes, channel, channel - channels[0])

    for info in build_info.values():
        if info.returncode != 0:
            return 1
    return 0


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('ms', help='Measurement set to image')
    parser.add_argument('output_dir', help='Output directory')
    parser.add_argument('--skip-build', action='store_true', help='Use existing image files')
    parser.add_argument('--stokes', default='IQUV', help='Stokes parameters to show')
    parser.add_argument('--start-channel', type=int, default=0, help='First channel to image')
    parser.add_argument('--stop-channel', type=int, default=1,
                        help='One past last channel to image')
    args = parser.parse_args()

    pixel_size = 1.7    # in arcsec
    pixels = 5760
    katsdpimager_common = [
        'imager.py',
        '--stokes=${stokes}',
        '--input-option', 'data=CORRECTED_DATA',
        '--eps-w=0.001', '--major=5',
        '--start-channel=${channels[0]}',
        '--stop-channel=${channels[-1] + 1}',
        '--pixel-size={}arcsec'.format(pixel_size),
        '--pixels={}'.format(pixels),
        '${ms}']
    lwimager_common = [
        'lwimager', 'ms=${ms}', 'npix={}'.format(pixels), 'mode=channel',
        'chanstart=${channels[0]}', 'nchan=${len(channels)}',
        'img_chanstart=${channels[0]}', 'img_nchan=${len(channels)}',
        'cellsize={}arcsec'.format(pixel_size), 'wprojplanes=128', 'threshold=0.01Jy',
        'weight=natural', 'stokes=${stokes}', 'data=CORRECTED_DATA']
    images = [
        Image('WSClean', 'wsclean',
              'wsclean-{rel_channel:04}-{stokes}-image.fits',
              'wsclean-{rel_channel:04}-{stokes}-dirty.fits',
              [['wsclean', '-mgain', '0.85', '-niter', '1000', '-threshold', '0.01',
                '-weight', 'natural',
                '-size', '{}'.format(pixels), '{}'.format(pixels),
                '-scale', '{}asec'.format(pixel_size), '-pol', '${",".join(stokes.lower())}',
                '-channels-out', '${len(channels)}',
                '-channel-range', '${channels[0]}', '${channels[-1] + 1}',
                '-name', '${output_dir}/wsclean', '${ms}']],
              clean_globs=['wsclean-*.fits']),
        Image('lwimager', 'lwimager', 'lwimager.fits', 'lwimager-dirty.fits',
              [
                  lwimager_common + ['operation=image', 'fits=${output_dir}/lwimager-dirty.fits'],
                  lwimager_common + ['operation=csclean', 'fits=${output_dir}/lwimager.fits']
              ],
              clean_globs=['*-channel*.img*']),
        Image('katsdpimager (GPU, direct predict)', 'katsdpimager-gpu',
              'image-gpu-{channel}.fits', 'dirty-gpu-{channel}.fits',
              [katsdpimager_common + ['--write-dirty=${output_dir}/dirty-gpu-%d.fits',
                                      '${output_dir}/image-gpu-%d.fits']]),
        Image('katsdpimager (GPU, degridding)', 'katsdpimager-gpu-degrid',
              'image-gpu-degrid-{channel}.fits', 'dirty-gpu-degrid-{channel}.fits',
              [katsdpimager_common + ['--degrid',
                                      '--write-dirty=${output_dir}/dirty-gpu-degrid-%d.fits',
                                      '${output_dir}/image-gpu-degrid-%d.fits']]),
        Image('katsdpimager (CPU)', 'katsdpimager-cpu',
              'image-cpu-{channel}.fits', 'dirty-cpu-{channel}.fits',
              [katsdpimager_common + ['--host', '--write-dirty=${output_dir}/dirty-cpu-%d.fits',
                                      '${output_dir}/image-cpu-%d.fits']])
    ]
    return run(args, images, MODES)


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    sys.exit(main())



# -*- coding: utf-8 -*-
"""Predict visibilities from a sky model by directly evaluating the RIME.

It uses a sky model to predict visibilities and subtract them from previous
values. No effects are modelled other than geometric delay (K Jones).
Direction-independent antenna effects (G Jones) should be already incorporated
into the visibilities which are being subtracted from, and time- and
antenna-independent DDEs should be included in the sky model. General DDEs are
not supported at all.

The visibilities are assumed to have already been passed through the
:mod:`preprocess` module, and hence UVW coordinates will be quantised.

.. include:: macros.rst
"""

import logging

import numpy as np
import pkg_resources
from astropy import units
from katsdpsigproc import accel, tune

from . import polarization, types, grid, parameters, numba


logger = logging.getLogger(__name__)


def _extract_sky_model(image_parameters, grid_parameters, model, phase_centre):
    """Extract lmn coordinates and fluxes from a sky model.

    Parameters
    ----------
    image_parameters : :class:`~.ImageParameters`
        Image parameters
    grid_parameters : :class:`~.GridParameters`
        Grid parameters for the corresponding UVW coordinates, for aliasing correction.
    model : :class:`~.SkyModel`
        The input sky model

    Returns
    -------
    lmn : array of float32
        Array of shape N×3, containing l, m, n-1
    flux : array of float32
        Array of shape N×P, containing the Stokes parameters specified by
        `image_parameters` for each source.
    """
    ip = image_parameters
    lmn = model.lmn(phase_centre)
    lmn[:, 2] -= 1    # n -> n-1
    flux = model.flux_density(ip.wavelength)
    # When we convert the subtracted visibilities back to a dirty image,
    # we compensate for the UV coordinate quantisation. However, in this case
    # we're already predicting with the quantised coordinates, so we need to
    # reverse that taper.
    #
    # I'm not sure that this is mathematically defensible, but the results
    # look good.
    taper = np.sinc(lmn[:, 0:2] / float(ip.image_size * grid_parameters.oversample))
    flux *= np.product(taper, axis=1, keepdims=True)

    # For each image polarization, find the corresponding index in IQUV
    # for advanced indexing.
    pol_index = [polarization.STOKES_IQUV.index(pol) for pol in ip.polarizations]
    flux = flux[:, pol_index]
    return lmn.astype(np.float32), flux.astype(np.float32)


def _extract_sky_image(image_parameters, grid_parameters, image):
    """Turn a sky model image into a model for direct prediction.

    The return values have the same meanings as for :func:`_extract_sky_model`.

    Parameters
    ----------
    image_parameters : :class:`~.ImageParameters`
        Image parameters
    grid_parameters : :class:`~.GridParameters`
        Grid parameters for the corresponding UVW coordinates, for aliasing correction.
    image : array of float
        P×size×size image

    Returns
    -------
    lmn : array of float32
        Array of shape N×3, containing l, m, n-1
    flux : array of float
        Array of shape N×P, containing the Stokes parameters specified by
        `image_parameters` for each source.
    """
    dtype = image.dtype
    pols = image.shape[2]
    # Create a mask of non-zero pixels
    mask = np.logical_or.reduce(image.astype(np.bool_), axis=0)
    pos = np.nonzero(mask)
    N = len(pos[0])

    lmn = np.empty((N, 3), np.float32)
    flux = np.empty((N, pols), dtype)
    # Note: These are currently done in double precision to avoid cancellation
    # issues in computing n-1.
    pixel_size = float(image_parameters.pixel_size)   # unitless Quantity -> plain float
    l = (pos[1] - 0.5 * image_parameters.pixels) * pixel_size
    m = (pos[0] - 0.5 * image_parameters.pixels) * pixel_size
    n1 = np.sqrt(1.0 - (np.square(l) + np.square(m))) - 1.0
    lmn[:, 0] = l
    lmn[:, 1] = m
    lmn[:, 2] = n1
    flux = image[:, pos[0], pos[1]].T
    # See :func:`_extract_sky_model` for explanation of this tapering
    taper_scale = float(image_parameters.image_size * grid_parameters.oversample)
    taper = np.sinc(l / taper_scale) * np.sinc(m / taper_scale)
    flux *= taper[:, np.newaxis]
    return lmn, flux


def _uvw_scale_bias(image_parameters, grid_parameters):
    r"""Compute scale and bias values to turn UVW indices back into coordinates.

    The :mod:`preprocess` module quantises UVW coordinates; the values computed
    by this function can be used to convert them back to coordinates in
    wavelengths. Given quantised UV with grid cell :math:`g` and subpixel `s`,
    and W plane of :math:`w_p` in a W slice with centre :math:`w_0` wavelengths,
    the coordinates in wavelengths are

    .. math::

       uv &= uv_{scale}(oversample\cdot g + s + 0.5)\\
       w &= w_0 + w_{scale}w_p + w_bias
    """
    ip = image_parameters
    gp = grid_parameters

    # uv is first computed in subpixels, hence need to / by oversample
    uv_scale = ip.cell_size / gp.oversample
    w_scale = gp.max_w / ((gp.w_slices - 0.5) * gp.w_planes)

    # The above give uvw in length, but we want it in wavelengths. We then
    # convert the (unitless, but possibly not scale-free) Quantity to a
    # plain float.
    uv_scale = float(uv_scale / ip.wavelength)
    w_scale = float(w_scale / ip.wavelength)
    w_bias = (0.5 - 0.5 * gp.w_planes) * w_scale
    return uv_scale, w_scale, w_bias


class PredictTemplate:
    """Predict visibilities from a sky model by directly evaluating the RIME.

    Parameters
    ----------
    context : |Context|
        Context for which kernels will be compiled
    real_dtype : {np.float32, np.float64}
        Data type for internal accumulation of values. This does not affect
        the buffer sizes.
    num_polarizations : int
        Number of polarizations
    tuning : dict, optional
        Kernel tuning parameters. The keys are:

        wgs
            Work group size
    """

    autotune_version = 0

    def __init__(self, context, real_dtype, num_polarizations, tuning=None):
        if tuning is None:
            tuning = self.autotune(context, real_dtype, num_polarizations)
        self.wgs = tuning['wgs']
        self.real_dtype = real_dtype
        self.num_polarizations = num_polarizations
        params = {
            'real_type': types.dtype_to_ctype(real_dtype),
            'wgs': self.wgs,
            'num_polarizations': self.num_polarizations
        }
        self.program = accel.build(
            context, 'imager_kernels/predict.mako', params,
            extra_dirs=[pkg_resources.resource_filename(__name__, '')])

    @classmethod
    @tune.autotuner(test={'wgs': 128})
    def autotune(cls, context, real_dtype, num_polarizations):
        queue = context.create_tuning_command_queue()
        num_vis = 1000000
        # Need at least as many sources as the workgroup size to achieve full
        # throughput.
        num_sources = 1024
        vis = accel.SVMArray(context, (num_vis, num_polarizations), dtype=np.complex64)
        uv = accel.SVMArray(context, (num_vis, 4), dtype=np.int16)
        w_plane = accel.SVMArray(context, (num_vis,), dtype=np.int16)
        weights = accel.SVMArray(context, (num_vis, num_polarizations), dtype=np.float32)
        lmn = accel.SVMArray(context, (num_sources, 3), dtype=np.float32)
        flux = accel.SVMArray(context, (num_sources, num_polarizations), dtype=np.float32)

        # The values don't make any difference to auto-tuning; they just affect
        # scale factors that have no impact on control flow.
        image_parameters = parameters.ImageParameters(
            1, 3, 0.21 * units.m, None, polarization.STOKES_IQUV[:num_polarizations],
            real_dtype, 1 * units.arcsec, 4096)
        grid_parameters = parameters.GridParameters(
            7.0, 8, 5, 5, 5, 1 * units.m, 64)

        def generate(wgs):
            template = cls(context, real_dtype, num_polarizations, {'wgs': wgs})
            fn = template.instantiate(queue, image_parameters, grid_parameters,
                                      num_vis, num_sources)
            fn.bind(vis=vis, uv=uv, w_plane=w_plane, weights=weights, lmn=lmn, flux=flux)
            fn.num_vis = num_vis
            fn._num_sources = num_sources   # Skip actually creating a sky model
            return tune.make_measure(queue, fn)

        return tune.autotune(generate, wgs=[64, 128, 256, 512, 1024])

    def instantiate(self, *args, **kwargs):
        return Predict(self, *args, **kwargs)


class Predict(grid.VisOperation):
    """Instantiation of :class:`PredictTemplate`.

    Before using a constructed instance, it's necessary to first

    1. Configure the visibilities, by setting :attr:`num_vis` and
       calling :meth:`set_coordinates`, :meth:`set_vis` and :meth:`set_w`.
    2. Configure the sources, by calling :meth:`set_sky_model` or
       :meth:`set_sky_image`.

    .. rubric:: Slots

    In addition to those specified in :class:`~.grid.VisOperation`:

    **lmn** : S×3, float32
        For each source, coordinates l, m, n-1
    **flux** : S×P, float32
        For each source, the perceived flux density (in Jy) per polarisation
    **weights** : N×P, float32
        The statistical weights associated with the visibilities. The input
        visibilities are assumed to be pre-weighted, and the predicted
        visibility is scaled by the weight before subtraction.

    These slots must all be backed by SVM-allocated memory.

    Parameters
    ----------
    template : :class:`PredictTemplate`
        The template for the operation
    command_queue : |CommandQueue|
        The command queue for the operation
    image_parameters : :class:`~.ImageParameters`
        Parameters that determine the UVW quantisation
    grid_parameters : :class:`~.GridParameters`
        Parameters that determine the UVW quantisation
    max_vis : int
        Maximum number of visibilities this instance can support
    max_sources : int
        Maximum number of sources this instance can support
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots
    """
    def __init__(self, template, command_queue, image_parameters, grid_parameters,
                 max_vis, max_sources, allocator=None):
        if len(image_parameters.polarizations) != template.num_polarizations:
            raise ValueError('Mismatch in number of polarizations')
        super().__init__(command_queue, template.num_polarizations, max_vis,
                         allocator)
        self.template = template
        pol_dim = accel.Dimension(template.num_polarizations, exact=True)
        sources_dim = max(1, max_sources)   # Cannot allocate 0-byte buffer
        self.slots['lmn'] = accel.IOSlot((sources_dim, accel.Dimension(3, exact=True)), np.float32)
        self.slots['flux'] = accel.IOSlot((sources_dim, pol_dim), np.float32)
        self.slots['weights'] = accel.IOSlot(
            (max_vis, accel.Dimension(template.num_polarizations, exact=True)), np.float32)
        self._kernel = self.template.program.get_kernel('predict')
        self._num_sources = 0
        self.max_sources = max_sources
        self.image_parameters = image_parameters
        self.grid_parameters = grid_parameters
        self._w = 0.0

    def set_sky_model(self, model, phase_centre):
        """Set the sky model.

        This copies data to the device, so if `model` is altered, a subsequent
        call is needed to update the device buffers.

        TODO: could use some optimisation
        """
        N = len(model)
        if N > self.max_sources:
            raise ValueError('too many sources ({} > {})'.format(N, self.max_sources))
        lmn, flux = _extract_sky_model(self.image_parameters, self.grid_parameters,
                                       model, phase_centre)
        self.buffer('lmn')[:N] = lmn
        self.buffer('flux')[:N] = flux
        self._num_sources = N

    def set_sky_image(self, image):
        """Set the sky model from a model image.

        This extracts components from the image, so if the image is altered a
        subsequent call is needed.
        """
        lmn, flux = _extract_sky_image(self.image_parameters, self.grid_parameters, image)
        N = len(lmn)
        if N > self.max_sources:
            raise ValueError('too many components ({} > {})'.format(N, self.max_sources))
        # TODO: have _extract_sky_image write directly to the buffer
        self.buffer('lmn')[:N] = lmn
        self.buffer('flux')[:N] = flux
        self._num_sources = N

    @property
    def num_sources(self):
        return self._num_sources

    def set_weights(self, weights):
        """Set statistical weights on visibilities.

        Before calling, set :attr:`num_vis`.
        """
        N = self.num_vis
        if len(weights) != N:
            raise ValueError('Lengths do not match')
        self.buffer('weights')[:N] = weights

    def set_w(self, w):
        """Set the W slice.

        The W plane coordinates set by :meth:`set_coordinates` are taken
        to lie within a slice centred at `w`, which is specified in
        wavelengths.
        """
        self._w = w

    def _run(self):
        if self.num_vis == 0 or self.num_sources == 0:
            return
        uv = self.buffer('uv')
        w_plane = self.buffer('w_plane')
        lmn = self.buffer('lmn')
        flux = self.buffer('flux')
        vis = self.buffer('vis')
        weights = self.buffer('weights')

        uv_scale, w_scale, w_bias = _uvw_scale_bias(self.image_parameters, self.grid_parameters)
        w_bias += self._w

        self.command_queue.enqueue_kernel(
            self._kernel,
            [
                vis.buffer,
                uv.buffer,
                w_plane.buffer,
                weights.buffer,
                lmn.buffer,
                flux.buffer,
                np.int32(self.num_vis),
                np.int32(self.num_sources),
                np.int32(self.grid_parameters.oversample), np.float32(uv_scale),
                np.float32(w_scale), np.float32(w_bias)
            ],
            global_size=(accel.roundup(self.num_vis, self.template.wgs),),
            local_size=(self.template.wgs,)
        )


@numba.jit(nopython=True)
def _predict_host(vis, uv, sub_uv, w_plane, weights, lmn, flux,
                  oversample, uv_scale, w_scale, w_bias, accum):
    N = vis.shape[0]
    S = lmn.shape[0]
    P = vis.shape[1]
    m2pi = np.complex64(-2j * np.pi)
    uvw = np.empty(3, np.float32)
    for i in range(N):
        uvw[0] = (uv[i, 0] * oversample + sub_uv[i, 0] + np.float32(0.5)) * uv_scale
        uvw[1] = (uv[i, 1] * oversample + sub_uv[i, 1] + np.float32(0.5)) * uv_scale
        uvw[2] = w_plane[i] * w_scale + w_bias
        accum[:] = 0
        for j in range(S):
            phase = np.dot(lmn[j], uvw)
            rot = np.exp(m2pi * phase)
            for p in range(P):
                accum[p] += rot * flux[j, p]
        accum *= weights[i]
        vis[i] -= accum


class PredictHost(grid.VisOperationHost):
    def __init__(self, image_parameters, grid_parameters):
        super().__init__()
        self.image_parameters = image_parameters
        self.grid_parameters = grid_parameters
        self.lmn = None
        self.flux = None
        self._w = 0

    @grid.VisOperationHost.num_vis.setter
    def num_vis(self, value):
        grid.VisOperationHost.num_vis.fset(self, value)
        self.weights = None

    def set_weights(self, weights):
        """Set statistical weights"""
        if len(weights) != self.num_vis:
            raise ValueError('Lengths do not match')
        self.weights = weights

    def set_w(self, w):
        self._w = w

    def set_sky_model(self, model, phase_centre):
        """Set the sky model.

        This copies data, so if `model` is altered, a subsequent
        call is needed to update the internal structures.
        """
        self.lmn, self.flux = _extract_sky_model(self.image_parameters,
                                                 self.grid_parameters,
                                                 model, phase_centre)

    def set_sky_image(self, image):
        """Set the sky model from an image.

        This copies data, so if `image` is altered, a subsequent
        call is needed to update the internal structures.
        """
        self.lmn, self.flux = _extract_sky_image(self.image_parameters,
                                                 self.grid_parameters,
                                                 image)

    def __call__(self):
        """Subtract predicted visibilities from existing values"""
        uv_scale, w_scale, w_bias = _uvw_scale_bias(self.image_parameters, self.grid_parameters)
        w_bias += self._w
        _predict_host(self.vis, self.uv, self.sub_uv, self.w_plane, self.weights,
                      self.lmn, self.flux,
                      np.float32(self.grid_parameters.oversample),
                      np.float32(uv_scale), np.float32(w_scale), np.float32(w_bias),
                      np.zeros(len(self.image_parameters.polarizations),
                               self.image_parameters.complex_dtype))



# -*- coding: utf-8 -*-
"""Load a local sky model from file and transfer it to a model image

At present the only file format supported is a katpoint catalogue file, but
other formats (e.g. Tigger) may be added later.
"""

import logging
import urllib

import numpy as np
from astropy import units
import katpoint


logger = logging.getLogger(__name__)


class NoSkyModelError(Exception):
    """Attempted to load a sky model for continuum subtraction but there isn't one"""
    pass


class SkyModel:
    """A sky model which can be used to generate images at specified frequencies.

    This is an abstract base class.
    """
    def __len__(self):
        """Get number of sources in catalogue"""
        raise NotImplementedError         # pragma: nocover

    @units.quantity_input(phase_centre=units.rad)
    def lmn(self, phase_centre):
        """Get direction cosine coordinates of the sources

        Parameters
        ----------
        phase_centre : Quantity
            RA and DEC of phase centre

        Returns
        -------
        array
            The l/m/n coordinates, shape N×3
        """
        raise NotImplementedError         # pragma: nocover

    @units.quantity_input(wavelength=units.m, equivalencies=units.spectral())
    def flux_density(self, wavelength):
        """Get flux densities for the sources at the given wavelength/frequency.

        Parameters
        ----------
        wavelength : Quantity
            Wavelength or frequency

        Returns
        -------
        array
            Flux densities in Jy, shape N×4 for Stokes IQUV
        """
        raise NotImplementedError         # pragma: nocover


class KatpointSkyModel(SkyModel):
    """Sky model created from a :class:`katpoint.Catalogue`."""

    def __init__(self, catalogue):
        self._catalogue = catalogue
        positions = units.Quantity(np.empty((len(catalogue), 2)),
                                   unit=units.deg)
        for i, source in enumerate(catalogue):
            positions[i] = units.Quantity(source.astrometric_radec(), unit=units.rad)
        self._positions = positions

    def __len__(self):
        return len(self._catalogue)

    @units.quantity_input(phase_centre=units.rad)
    def lmn(self, phase_centre):
        phase_centre = phase_centre.to(units.rad).value
        phase_centre = katpoint.construct_radec_target(phase_centre[0], phase_centre[1])
        return np.array([phase_centre.lmn(*source.astrometric_radec())
                         for source in self._catalogue])

    @units.quantity_input(wavelength=units.m, equivalencies=units.spectral())
    def flux_density(self, wavelength):
        freq_MHz = wavelength.to(units.MHz, equivalencies=units.spectral()).value
        out = np.stack([source.flux_density_stokes(freq_MHz) for source in self._catalogue])
        return np.nan_to_num(out, copy=False)


def catalogue_from_telstate(telstate, capture_block_id, continuum, target):
    """Extract a katpoint catalogue written by katsdpcontim.

    Parameters
    ----------
    telstate : :class:`katsdptelstate.TelescopeState` or :class:`katdal.sensordata.TelstateToStr`
        Telescope state
    capture_block_id : str
        Capture block ID
    continuum : str or ``None``
        Name of the continuum imaging stream (used to form the telstate view).
        If ``None``, there must be exactly one continuum imaging stream in the
        data set, which is used.
    target : :class:`katpoint.Target`
        Target field

    Raises
    ------
    NoSkyModelError
        If no sky model could be found for the given parameters

    Returns
    -------
    katpoint.Catalogue
    """
    # katsdpcontim is still Python 2, so it needs TelstateToStr to work with
    # if we are run under Python 3.
    from katdal.sensordata import TelstateToStr

    telstate = TelstateToStr(telstate)
    try:
        # Find the continuum image stream
        if continuum is None:
            archived_streams = telstate['sdp_archived_streams']
            for stream_name in archived_streams:
                view = telstate.view(stream_name, exclusive=True)
                view = view.view(telstate.join(capture_block_id, stream_name))
                stream_type = view.get('stream_type', 'unknown')
                if stream_type != 'sdp.continuum_image':
                    continue
                if continuum is not None:
                    raise NoSkyModelError(
                        'Multiple continuum image streams found - need to select one')
                continuum = stream_name
            if continuum is None:
                raise NoSkyModelError('No continuum image streams found')

        view = telstate.view(continuum, exclusive=True)
        view = view.view(telstate.join(capture_block_id, continuum))
        target_namespace = view['targets'][target.description]
        prefix = telstate.join(capture_block_id, continuum, target_namespace, 'target0')
        data = view.view(prefix)['clean_components']
        # Should always match, but for safety
        if katpoint.Target(data['description']) == target:
            return katpoint.Catalogue(data['components'])
    except KeyError:
        logger.debug('KeyError', exc_info=True)
    raise NoSkyModelError('Sky model for target {} not found'.format(target.name))


def open_sky_model(url):
    """Load a sky model from an external resource.

    The format is encoded in the URL as a `format` query parameter, and
    defaults to ``katpoint``.

    Parameters
    ----------
    url : str
        The interpretation depends on the format.

        katpoint
            A ``file://`` URL for a katpoint catalogue file
        katdal
            A katdal MVFv4 URL. The following extra query parameters are interpreted
            and removed before they are passed to katdal:

            target (required)
                Description of the katpoint target whose LSM should be loaded
            continuum (optional)
                Name of the continuum image stream. If not specified, there
                must be exactly one in the file.

    Raises
    ------
    ValueError
        if `format` was not recognised, the URL doesn't contain the
        expected query parameters, or the URL scheme is not supported
    IOError, OSError
        if there was a low-level error reading a file
    Exception
        any exception raised by katdal in opening the file

    Returns
    -------
    SkyModel
    """
    parts = urllib.parse.urlparse(url, scheme='file')
    params = urllib.parse.parse_qs(parts.query)
    model_format = params.pop('format', ['katpoint'])[0]

    if model_format == 'katdal':
        import katdal
        try:
            target = katpoint.Target(params.pop('target')[0])
        except KeyError:
            raise ValueError('URL must contain target')
        try:
            continuum = params.pop('continuum')[0]
        except KeyError:
            continuum = None
        # Reconstruct the URL without the query components we've absorbed
        new_url = urllib.parse.urlunparse((
            parts.scheme, parts.netloc, parts.path, parts.params,
            urllib.parse.urlencode(params, doseq=True),
            parts.fragment))
        data_source = katdal.open_data_source(new_url, chunk_store=None, upgrade_flags=False)
        catalogue = catalogue_from_telstate(data_source.telstate, data_source.capture_block_id,
                                            continuum, target)
        return KatpointSkyModel(catalogue)
    elif model_format == 'katpoint':
        if parts.scheme != 'file':
            raise ValueError('Only file:// URLs are supported for katpoint sky model format')
        with open(parts.path) as f:
            catalogue = katpoint.Catalogue(f)
        return KatpointSkyModel(catalogue)
    else:
        raise ValueError('format "{}" is not recognised'.format(model_format))



"""Tests for :py:mod:`katsdpimager.polarization`."""

import numpy as np
from nose.tools import assert_equal, assert_raises

from .. import polarization


class TestPolarizationMatrix:
    """Tests for :py:func:`katsdpimager.polarization.polarization_matrix`,
    using standard coordinate systems.
    """
    def setup(self):
        self.IQUV = polarization.STOKES_IQUV
        self.IQ = [polarization.STOKES_I, polarization.STOKES_Q]
        self.XY = [polarization.STOKES_XX,
                   polarization.STOKES_XY,
                   polarization.STOKES_YX,
                   polarization.STOKES_YY]
        self.XY_DIAG = [polarization.STOKES_XX, polarization.STOKES_YY]
        self.RL = [polarization.STOKES_RR,
                   polarization.STOKES_RL,
                   polarization.STOKES_LR,
                   polarization.STOKES_LL]
        self.RL_DIAG = [polarization.STOKES_RR, polarization.STOKES_LL]

    def test_xy_to_iquv(self):
        expected = 0.5 * np.matrix([[1, 0, 0, 1],
                                    [1, 0, 0, -1],
                                    [0, 1, 1, 0],
                                    [0, -1j, 1j, 0]])
        actual = polarization.polarization_matrix(self.IQUV, self.XY)
        np.testing.assert_array_equal(expected, actual)

    def test_stokes_to_rl(self):
        expected = np.matrix([[1, 0, 0, 1],
                              [0, 1, 1j, 0],
                              [0, 1, -1j, 0],
                              [1, 0, 0, -1]])
        actual = polarization.polarization_matrix(self.RL, self.IQUV)
        np.testing.assert_array_equal(expected, actual)

    def test_xy_to_rl(self):
        expected = 0.5 * np.matrix([[1, -1j, 1j, 1],
                                    [1, 1j, 1j, -1],
                                    [1, -1j, -1j, -1],
                                    [1, 1j, -1j, 1]])
        actual = polarization.polarization_matrix(self.RL, self.XY)
        np.testing.assert_array_equal(expected, actual)

    def test_xy_diag_to_iq(self):
        expected = 0.5 * np.matrix([[1, 1], [1, -1]])
        actual = polarization.polarization_matrix(self.IQ, self.XY_DIAG)
        np.testing.assert_array_equal(expected, actual)

    def test_xy_diag_to_i(self):
        expected = 0.5 * np.matrix([[1, 1]])
        actual = polarization.polarization_matrix([polarization.STOKES_I], self.XY_DIAG)
        np.testing.assert_array_equal(expected, actual)

    def test_xy_diag_to_iquv(self):
        with assert_raises(ValueError):
            polarization.polarization_matrix(self.IQUV, self.XY_DIAG)


class TestParseStokes:
    def test_good(self):
        assert_equal(polarization.parse_stokes('QIV'),
                     [polarization.STOKES_I, polarization.STOKES_Q, polarization.STOKES_V])

    def test_empty(self):
        with assert_raises(ValueError):
            polarization.parse_stokes('')

    def test_duplicate(self):
        with assert_raises(ValueError):
            polarization.parse_stokes('QQ')

    def test_invalid(self):
        with assert_raises(ValueError):
            polarization.parse_stokes('Z')



#!/usr/bin/env python3
import argparse
import logging
import os
import uuid
import shutil
import json
from datetime import datetime
from contextlib import closing
import gc

from astropy import units
from astropy.coordinates import Angle
import katsdpservices
import katdal
from katsdpsigproc import accel

from katsdpimager import frontend, loader, io, progress, render


logger = logging.getLogger()


class Writer(frontend.Writer):
    def __init__(self, args, dataset):
        self.args = args
        self.uuid = uuid.uuid4()

        raw_data = dataset.raw_data
        if not isinstance(raw_data, katdal.DataSet):
            raise RuntimeError('Only katdal data sets are supported')
        radec_deg = dataset.phase_centre().to(units.deg)
        obs_params = raw_data.obs_params
        self.common_metadata = {
            'ProductType': {
                'ProductTypeName': 'FITSImageProduct',
                'ReductionName': 'Spectral Image'
            },
            'CaptureBlockId': raw_data.source.capture_block_id,
            'ScheduleBlockIdCode': obs_params.get('sb_id_code', 'UNKNOWN'),
            'Description': obs_params.get('description', 'UNKNOWN') + ': Spectral-line image',
            'ProposalId': obs_params.get('proposal_id', 'UNKNOWN'),
            'Observer': obs_params.get('observer', 'UNKNOWN'),
            # Solr doesn't accept +00:00, only Z, so we can't just format a timezone-aware value
            'StartTime': datetime.utcnow().isoformat() + 'Z',
            'Bandwidth': raw_data.channel_width,
            'ChannelWidth': raw_data.channel_width,
            'NumFreqChannels': 1,
            'RightAscension': [Angle(radec_deg[0]).to_string(units.hour, sep=':', pad=True)],
            'Declination': [Angle(radec_deg[1]).to_string(units.deg, sep=':', pad=True)],
            # JSON schema limits format to fixed-point with at most 10 decimal places
            'DecRa': [','.join('{:.10f}'.format(angle) for angle in radec_deg.value[::-1])],
            'Targets': [dataset.raw_target.name],
            'KatpointTargets': [dataset.raw_target.description],
            # katdal gives dump period in seconds, metadata value needs to be in hours
            'IntegrationTime': [raw_data.dump_period * len(raw_data.dumps) / 3600.0]
        }

    def write_fits_image(self, name, description, dataset, image, image_parameters, channel,
                         beam=None, bunit='JY/BEAM'):
        if name != 'clean':
            return
        # Add a unique component to the directory name so that it will be
        # unique even if the task dies and is re-run.
        output_dir = '{}_{:05}_{}'.format(self.args.prefix, channel, self.uuid)
        output_dir = os.path.join(self.args.output_dir, output_dir)
        tmp_dir = output_dir + '.writing'
        base_filename = '{}_{:05}.fits'.format(self.args.prefix, channel)
        filename = os.path.join(tmp_dir, base_filename)
        freq = image_parameters.wavelength.to(units.Hz, equivalencies=units.spectral()).value
        channel_width = dataset.raw_data.channel_width
        metadata = {
            **self.common_metadata,
            'FITSImageFilename': [base_filename],
            'PNGImageFileName': [base_filename + '.png'],
            'PNGThumbNailFileName': [base_filename + '.tnail.png'],
            'CenterFrequency': freq,
            'MinFreq': freq - 0.5 * channel_width,
            'MaxFreq': freq + 0.5 * channel_width,
            'Run': channel
        }
        os.mkdir(tmp_dir)
        try:
            with progress.step('Write {}'.format(description)):
                io.write_fits_image(dataset, image, image_parameters, filename,
                                    channel, beam, bunit)
            with progress.step('Write PNG'):
                render.write_image(filename, filename + '.png',
                                   width=6500, height=5000,
                                   dpi=10 * render.DEFAULT_DPI)
            with progress.step('Write thumbnail'):
                render.write_image(filename, filename + '.tnail.png', width=650, height=500)
            with open(os.path.join(tmp_dir, 'metadata.json'), 'w') as f:
                json.dump(metadata, f, allow_nan=False, indent=2)
            os.rename(tmp_dir, output_dir)
        except Exception:
            # Make a best effort to clean up
            shutil.rmtree(tmp_dir, ignore_errors=True)
            raise
        finally:
            # Something in the PNG writing causes memory to not get freed,
            # and the garbage collector is not kicking in properly on its
            # own.
            gc.collect()

    def write_fits_grid(self, name, description, fftshift, grid_data, image_parameters, channel):
        pass


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('input_file', type=str, metavar='INPUT',
                        help='Input data set')
    parser.add_argument('output_dir', type=str, metavar='DIR',
                        help='Parent directory for output')
    parser.add_argument('prefix', type=str,
                        help='Prefix for output directories and filenames')
    parser.add_argument('--log-level', type=str, metavar='LEVEL',
                        help='Logging level [INFO]')

    frontend.add_options(parser)

    # Arguments only supported by offline imager.py
    parser.set_defaults(host=False, vis_limit=None)
    return parser


def main():
    parser = get_parser()
    args = parser.parse_args()
    args.input_option = ['--' + opt for opt in args.input_option]
    katsdpservices.setup_logging()
    if args.log_level is not None:
        logger.setLevel(args.log_level.upper())

    with closing(loader.load(args.input_file, args.input_option)) as dataset:
        writer = Writer(args, dataset)
        context = accel.create_some_context(interactive=False, device_filter=lambda x: x.is_cuda)
        queue = context.create_command_queue()
        frontend.run(args, context, queue, dataset, writer)


if __name__ == '__main__':
    main()



"""Kernels for image-domain processing

It also handles conversion between visibility and image planes.

.. include:: macros.rst
"""

import math

import numpy as np
import pkg_resources
from katsdpsigproc import accel

import katsdpimager.types
from . import fft


class _LayerImageTemplate:
    r"""Base class for :class:`LayerToImageTemplate` and :class:`ImageToLayerTemplate`.

    These operations convert between a "layer" (the raw Fourier Transform from
    the UV plane) and the "image". :class:`LayerToImageTemplate` converts from
    layer to image (accumulating into the stacked image), and
    :class:`ImageToLayerTemplate` is the inverse. :class:`LayerToImageTemplate`
    applies the following operations:

    - It reorders the elements to put the image centre at the centre (in the layer,
      it is in the corners). This is the equivalent of `np.fft.fftshift`.

    - It divides out an image tapering function. The function is the
      combination of a separable antialiasing function and the 3rd direction
      cosine (n). Specifically, for input pixel coordinates x, y, we
      calculate

      .. math::

         l(x) &= \text{lm_scale}\cdot x + \text{lm_bias}\\
         m(y) &= \text{lm_scale}\cdot y + \text{lm_bias}\\
         f(x, y) &= \frac{\text{kernel1d}[x] \cdot \text{kernel1d}[y]}{\sqrt{1-l(x)^2-m(y)^2}}

      This function is also *divided* in the reverse direction, making the two
      operations not actually inverses of each other. This is necessary because
      the transform of the tapering function is convolved with visibilities in
      both directions.

    - It multiplies by a W correction term, namely

      .. math::

         e^{2\pi i w\left(\sqrt{1-l(x)^2-m(y)^2} - 1\right)}

         Only the real component of the result is kept in the image.

    Further dimensions are supported e.g. for Stokes parameters, which occur
    before the l/m dimensions.

    A note on sign conventions: the measurement equation is taken to be

    .. math::

       V(u, v, w) = \int \frac{I(l, m)}{n} e^{-2\pi i(ul + vm + w(n-1))}\ dl\ dm.

    This is consistent with the sign conventions for the phase of :math:`V` and
    for :math:`(u, v, w)` that are documented in
    :ref:`katsdpimager.loader_core.LoaderBase.data_iter`.

    Parameters
    ----------
    context : |Context|
        Context for which kernels will be compiled
    real_dtype : {`np.float32`, `np.float64`}
        Image type
    kernel_filename : str
        Mako template containing the kernel
    tuning : dict, optional
        Tuning parameters (currently unused)
    """
    def __init__(self, context, real_dtype, kernel_filename, tuning=None):
        self.real_dtype = np.dtype(real_dtype)
        self.wgs_x = 16  # TODO: autotuning
        self.wgs_y = 16
        self.program = accel.build(
            context, kernel_filename,
            {
                'real_type': katsdpimager.types.dtype_to_ctype(real_dtype),
                'wgs_x': self.wgs_x,
                'wgs_y': self.wgs_y,
            },
            extra_dirs=[pkg_resources.resource_filename(__name__, '')])


class _LayerImage(accel.Operation):
    """Base class for instantiation of subclasses of :class:`_LayerImageTemplate`

    .. rubric:: Slots

    **layer** : array of complex values
        Input (for divide) / output (for multiply)
    **image** : array of real values
        Output (for divide) / input (for multiply)
    **kernel1d** : array of real values, 1D
        Antialiasing function

    Parameters
    ----------
    template : :class:`_LayerImageTemplate`
        Operation template
    command_queue : |CommandQueue|
        Command queue for the operation
    shape : tuple of int
        Shape of the image data (polarizations, height, width) - must be square
    lm_scale : float
        Scale factor from pixel coordinates to l/m coordinates
    lm_bias : float
        Bias from scaled pixel coordinates to l/m coordinates
    kernel_name : str
        Name of the kernel function
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots

    Raises
    ------
    ValueError
        if the last two elements of shape are not equal and even
    """
    def __init__(self, template, command_queue, shape, lm_scale, lm_bias, kernel_name,
                 allocator=None):
        if len(shape) != 3 or shape[-1] != shape[-2]:
            raise ValueError('shape must be square, not {}'.format(shape))
        if shape[-1] % 2 != 0:
            raise ValueError('image size must be even, not {}'.format(shape[-1]))
        super().__init__(command_queue, allocator)
        self.template = template
        complex_dtype = katsdpimager.types.real_to_complex(template.real_dtype)
        dims = [accel.Dimension(x) for x in shape]
        self.slots['layer'] = accel.IOSlot(dims[-2:], complex_dtype)
        self.slots['image'] = accel.IOSlot(dims, template.real_dtype)
        self.slots['kernel1d'] = accel.IOSlot((shape[-1],), template.real_dtype)
        self.kernel = template.program.get_kernel(kernel_name)
        self.lm_scale = lm_scale
        self.lm_bias = lm_bias
        self.w = 0
        self.polarization = 0

    def set_w(self, w):
        """Set w (in wavelengths)."""
        self.w = w

    def set_polarization(self, polarization):
        """Set polarization index in the image"""
        if polarization < 0 or polarization >= self.slots['image'].shape[0]:
            raise IndexError('polarization index out of range')
        self.polarization = polarization

    def _run(self):
        layer = self.buffer('layer')
        image = self.buffer('image')
        assert layer.padded_shape == image.padded_shape[-2:]
        half_size = image.shape[-1] // 2
        row_stride = image.padded_shape[-1]
        slice_stride = row_stride * image.padded_shape[-2]
        kernel1d = self.buffer('kernel1d')
        self.command_queue.enqueue_kernel(
            self.kernel,
            [
                image.buffer,
                layer.buffer,
                np.int32(self.polarization * slice_stride),
                np.int32(row_stride),
                kernel1d.buffer,
                self.template.real_dtype.type(self.lm_scale),
                self.template.real_dtype.type(self.lm_bias),
                np.int32(half_size),
                np.int32(half_size * row_stride),
                self.template.real_dtype.type(half_size * self.lm_scale),
                self.template.real_dtype.type(2 * self.w)
            ],
            global_size=(accel.roundup(half_size, self.template.wgs_x),
                         accel.roundup(half_size, self.template.wgs_y)),
            local_size=(self.template.wgs_x, self.template.wgs_y)
        )


class LayerToImageTemplate(_LayerImageTemplate):
    """Convert layer to image. See :class:`_LayerImageTemplate` for details.

    Parameters
    ----------
    context : |Context|
        Context for which kernels will be compiled
    real_dtype : {`np.float32`, `np.float64`}
        Image type
    tuning : dict, optional
        Tuning parameters (currently unused)
    """
    def __init__(self, context, real_dtype, tuning=None):
        super().__init__(
            context, real_dtype, 'imager_kernels/layer_to_image.mako', tuning)

    def instantiate(self, *args, **kwargs):
        return LayerToImage(self, *args, **kwargs)


class LayerToImage(_LayerImage):
    """Instantiation of :class:`LayerToImageTemplate`. See :class:`_LayerImage` for
    details.

    Parameters
    ----------
    template : :class:`LayerToImageTemplate`
        Operation template
    command_queue : |CommandQueue|
        Command queue for the operation
    shape : tuple of int
        Shape of the data (must be square)
    lm_scale : float
        Scale factor from pixel coordinates to l/m coordinates
    lm_bias : float
        Bias from scaled pixel coordinates to l/m coordinates
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots

    Raises
    ------
    ValueError
        if the last two elements of shape are not equal and even
    """
    def __init__(self, template, command_queue, shape, lm_scale, lm_bias, allocator=None):
        super().__init__(
            template, command_queue, shape, lm_scale, lm_bias, "layer_to_image", allocator)


class ImageToLayerTemplate(_LayerImageTemplate):
    """Convert image to layer. See :class:`_LayerImageTemplate` for details.

    Parameters
    ----------
    context : |Context|
        Context for which kernels will be compiled
    real_dtype : {`np.float32`, `np.float64`}
        Image type
    tuning : dict, optional
        Tuning parameters (currently unused)
    """
    def __init__(self, context, real_dtype, tuning=None):
        super().__init__(
            context, real_dtype, 'imager_kernels/image_to_layer.mako', tuning)

    def instantiate(self, *args, **kwargs):
        return ImageToLayer(self, *args, **kwargs)


class ImageToLayer(_LayerImage):
    """Instantiation of :class:`ImageToLayerTemplate`. See :class:`_LayerImage` for
    details.

    Parameters
    ----------
    template : :class:`ImageToLayerTemplate`
        Operation template
    command_queue : |CommandQueue|
        Command queue for the operation
    shape : tuple of int
        Shape of the image as (polarizations, height, width) - must be square
    lm_scale : float
        Scale factor from pixel coordinates to l/m coordinates
    lm_bias : float
        Bias from scaled pixel coordinates to l/m coordinates
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots

    Raises
    ------
    ValueError
        if the last two elements of shape are not equal and even
    """
    def __init__(self, template, command_queue, shape, lm_scale, lm_bias, allocator=None):
        super().__init__(
            template, command_queue, shape, lm_scale, lm_bias, "image_to_layer", allocator)


class ScaleTemplate:
    """Scale an image by a fixed amount per polarization.

    Parameters
    ----------
    context : |Context|
        Context for which kernels will be compiled
    dtype : {`np.float32`, `np.float64`}
        Image precision
    num_polarizations : int
        Number of polarizations stored in the image
    tuning : dict, optional
        Tuning parameters (currently unused)
    """

    def __init__(self, context, dtype, num_polarizations, tuning=None):
        # TODO: autotuning
        self.context = context
        self.dtype = np.dtype(dtype)
        self.num_polarizations = num_polarizations
        self.wgsx = 16
        self.wgsy = 16
        self.program = accel.build(
            context, "imager_kernels/scale.mako",
            {
                'real_type': katsdpimager.types.dtype_to_ctype(dtype),
                'wgsx': self.wgsx,
                'wgsy': self.wgsy,
                'num_polarizations': num_polarizations
            },
            extra_dirs=[pkg_resources.resource_filename(__name__, '')])

    def instantiate(self, *args, **kwargs):
        return Scale(self, *args, **kwargs)


class Scale(accel.Operation):
    """Instantiation of :class:`ScaleTemplate`.

    .. rubric:: Slots

    **data** : array
        Image, indexed by polarization, y, x

    Parameters
    ----------
    template : :class:`ScaleTemplate`
        Operation template
    command_queue : |CommandQueue|
        Command queue for the operation
    shape : tuple of int
        Shape of the data.
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots
    """
    def __init__(self, template, command_queue, shape, allocator=None):
        super().__init__(command_queue, allocator)
        self.template = template
        if len(shape) != 3:
            raise ValueError('Wrong number of dimensions in shape')
        if shape[0] != template.num_polarizations:
            raise ValueError('Mismatch in number of polarizations')
        self.slots['data'] = accel.IOSlot(shape, template.dtype)
        self.kernel = template.program.get_kernel('scale')
        self.scale_factor = np.zeros((shape[0],), template.dtype)

    def set_scale_factor(self, scale_factor):
        self.scale_factor[:] = scale_factor

    def _run(self):
        data = self.buffer('data')
        self.command_queue.enqueue_kernel(
            self.kernel,
            [
                data.buffer,
                np.int32(data.padded_shape[2]),
                np.int32(data.padded_shape[1] * data.padded_shape[2]),
                np.int32(data.shape[2]),
                np.int32(data.shape[1]),
                self.scale_factor
            ],
            global_size=(accel.roundup(data.shape[2], self.template.wgsx),
                         accel.roundup(data.shape[1], self.template.wgsy)),
            local_size=(self.template.wgsx, self.template.wgsy)
        )


class GridImageTemplate:
    """Template for a combined operation that converts from a complex grid to a
    real image or vice versa, including layer-to-image/image-to-layer
    conversion.  The grid need not be conjugate symmetric: it is put through a
    complex-to-complex transformation, and the real part of the result is
    returned. Both the grid and the image have the DC term in the middle.

    Because it uses :class:`~katsdpimager.fft.FftTemplate`, most of the
    parameters are baked into the template rather than the instance.

    .. warning::
        Instances created from the same template cannot be executed
        simultaneously, due to limitations of CUFFT (work area is part of the
        plan).

    Parameters
    ----------
    command_queue : :class:`katsdpsigproc.cuda.CommandQueue`
        Command queue for the operation
    shape_layer : tuple of int
        Shape for the intermediate layer, as (height, width)
    padded_shape_layer : tuple of int
        Padded shape of the intermediate layer
    real_dtype : {`np.float32`, `np.float64`}
        Precision
    """

    def __init__(self, command_queue, shape_layer, padded_shape_layer, real_dtype):
        complex_dtype = katsdpimager.types.real_to_complex(real_dtype)
        self.fft = fft.FftTemplate(command_queue, 2, shape_layer, complex_dtype, complex_dtype,
                                   padded_shape_layer, padded_shape_layer)
        self.layer_to_image = LayerToImageTemplate(command_queue.context, real_dtype)
        self.image_to_layer = ImageToLayerTemplate(command_queue.context, real_dtype)

    def instantiate_grid_to_image(self, *args, **kwargs):
        return GridToImage(self, *args, **kwargs)

    def instantiate_image_to_grid(self, *args, **kwargs):
        return ImageToGrid(self, *args, **kwargs)


class GridToImage(accel.OperationSequence):
    """Instantiation of :class:`GridToImageTemplate`.

    Parameters
    ----------
    template : :class:`GridImageTemplate`
        Operation template
    shape_grid : tuple of int
        Shape of the grid, as (polarizations, height, width)
    lm_scale,lm_bias : float
        See :class:`LayerToImage`
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots
    """
    def __init__(self, template, shape_grid, lm_scale, lm_bias, allocator=None):
        command_queue = template.fft.command_queue
        self._ifft = template.fft.instantiate(fft.FFT_INVERSE, allocator)
        polarizations = shape_grid[0]
        shape_image = (polarizations,) + tuple(template.fft.shape)
        self._layer_to_image = template.layer_to_image.instantiate(
            command_queue, shape_image, lm_scale, lm_bias, allocator)
        operations = [
            ('ifft', self._ifft),
            ('layer_to_image', self._layer_to_image)
        ]
        compounds = {
            'layer': ['ifft:src', 'ifft:dest', 'layer_to_image:layer'],
            'image': ['layer_to_image:image'],
            'kernel1d': ['layer_to_image:kernel1d']
        }
        super().__init__(command_queue, operations, compounds, allocator=allocator)
        self.slots['grid'] = accel.IOSlot(shape_grid, template.fft.dtype_src)

    def set_w(self, w):
        self._layer_to_image.set_w(w)

    def _run(self):
        grid = self.buffer('grid')
        layer = self.buffer('layer')
        polarizations, width, height = grid.shape
        # This could probably be made to work correctly with odd width/height,
        # but it would need to be done carefully to avoid off-by-one bugs.
        assert width % 2 == 0
        assert height % 2 == 0
        half_width = width // 2
        half_height = height // 2
        for pol in range(polarizations):
            layer.zero(self.command_queue)
            # Copy one polarization from the grid to the layer, also switching
            # the centre to the corners. The two slices in each of src/dest_x/y
            # are the two halves of the data.
            src_y = np.s_[:half_height, -half_height:]
            dest_y = np.s_[-half_height:, :half_height]
            src_x = np.s_[:half_width, -half_width:]
            dest_x = np.s_[-half_width:, :half_width]
            for sy, dy in zip(src_y, dest_y):
                for sx, dx in zip(src_x, dest_x):
                    grid.copy_region(
                        self.command_queue, layer, (pol, sy, sx), (dy, dx))
            self._layer_to_image.set_polarization(pol)
            super()._run()


class ImageToGrid(accel.OperationSequence):
    """Instantiation of :class:`ImageToGridTemplate`.

    Parameters
    ----------
    template : :class:`GridImageTemplate`
        Operation template
    shape_grid : tuple of int
        Shape of the grid, as (polarizations, height, width)
    lm_scale,lm_bias : float
        See :class:`ImageToLayer`
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots
    """
    def __init__(self, template, shape_grid, lm_scale, lm_bias, allocator=None):
        command_queue = template.fft.command_queue
        polarizations = shape_grid[0]
        shape_image = (polarizations,) + tuple(template.fft.shape)
        self._fft = template.fft.instantiate(fft.FFT_FORWARD, allocator)
        self._image_to_layer = template.image_to_layer.instantiate(
            command_queue, shape_image, lm_scale, lm_bias, allocator)
        operations = [
            ('image_to_layer', self._image_to_layer),
            ('fft', self._fft),
        ]
        compounds = {
            'layer': ['fft:src', 'fft:dest', 'image_to_layer:layer'],
            'image': ['image_to_layer:image'],
            'kernel1d': ['image_to_layer:kernel1d']
        }
        super().__init__(command_queue, operations, compounds, allocator=allocator)
        self.slots['grid'] = accel.IOSlot(shape_grid, template.fft.dtype_dest)

    def set_w(self, w):
        self._image_to_layer.set_w(w)

    def _run(self):
        # TODO: unify with GridToImage._run
        grid = self.buffer('grid')
        layer = self.buffer('layer')
        polarizations, width, height = grid.shape
        # This could probably be made to work correctly with odd width/height,
        # but it would need to be done carefully to avoid off-by-one bugs.
        assert width % 2 == 0
        assert height % 2 == 0
        half_width = width // 2
        half_height = height // 2
        for pol in range(polarizations):
            self._image_to_layer.set_polarization(pol)
            super()._run()
            # Copy one polarization from the grid to the layer, also switching
            # the centre to the corners. The two slices in each of src/dest_x/y
            # are the two halves of the data.
            src_y = np.s_[:half_height, -half_height:]
            dest_y = np.s_[-half_height:, :half_height]
            src_x = np.s_[:half_width, -half_width:]
            dest_x = np.s_[-half_width:, :half_width]
            for sy, dy in zip(src_y, dest_y):
                for sx, dx in zip(src_x, dest_x):
                    layer.copy_region(
                        self.command_queue, grid, (sy, sx), (pol, dy, dx))


class GridToImageHost:
    """CPU-only equivalent to :class:`GridToImage`.

    The parameters specify which buffers the operation runs on, but the
    contents at construction time are irrelevant. The operation is
    performed at call time.

    Parameters
    ----------
    grid : ndarray, complex
        Input grid (unmodified)
    layer : ndarray, complex
        Intermediate structure holding the complex FFT
    image : ndarray, real
        Output image
    kernel1d : ndarray, real
        Tapering function
    lm_scale,lm_bias : real
        Linear transformation from pixel coordinates to l/m values
    """

    def __init__(self, grid, layer, image, kernel1d, lm_scale, lm_bias):
        assert image.shape[-1] == image.shape[-2]
        assert image.shape[-1] % 2 == 0
        self.grid = grid
        self.layer = layer
        self.image = image
        self.kernel1d = kernel1d
        self.lm_scale = lm_scale
        self.lm_bias = lm_bias
        self.w = 0

    def set_w(self, w):
        self.w = w

    def clear(self):
        self.image.fill(0)

    def __call__(self):
        self.layer[:] = np.fft.ifft2(np.fft.ifftshift(self.grid, axes=(1, 2)), axes=(1, 2))
        # Scale factor is to match behaviour of CUFFT, which is unnormalized
        scale = self.layer.shape[1] * self.layer.shape[2]
        lm = np.arange(self.image.shape[1]).astype(self.image.dtype) * self.lm_scale + self.lm_bias
        # We do calculations involving the complex values prior to applying
        # fftshift to save memory, so we have to apply the inverse shift to lm.
        lm = np.fft.ifftshift(lm)
        lm2 = lm * lm
        n = np.sqrt(1 - (lm2[:, np.newaxis] + lm2[np.newaxis, :]))
        w_correct = np.exp(2j * math.pi * self.w * (n - 1))
        self.layer *= w_correct
        # TODO: most of the calculations here would be more efficient with numba
        image = self.layer.real.copy()
        image *= scale
        image *= n[np.newaxis, ...]
        image = np.fft.fftshift(image, axes=(1, 2))
        image /= np.outer(self.kernel1d, self.kernel1d)[np.newaxis, ...]
        self.image += image


class ImageToGridHost:
    """CPU-only equivalent to :class:`ImageToGrid`.

    The parameters specify which buffers the operation runs on, but the
    contents at construction time are irrelevant. The operation is
    performed at call time.

    Parameters
    ----------
    grid : ndarray, complex
        Output grid
    layer : ndarray, complex
        Intermediate structure holding the complex FFT
    image : ndarray, real
        Input image
    kernel1d : ndarray, real
        Tapering function
    lm_scale,lm_bias : real
        Linear transformation from pixel coordinates to l/m values
    """
    def __init__(self, grid, layer, image, kernel1d, lm_scale, lm_bias):
        assert image.shape[-1] == image.shape[-2]
        assert image.shape[-1] % 2 == 0
        self.grid = grid
        self.layer = layer
        self.image = image
        self.kernel1d = kernel1d
        self.lm_scale = lm_scale
        self.lm_bias = lm_bias
        self.w = 0

    def set_w(self, w):
        self.w = w

    def __call__(self):
        # TODO: rewrite most of this using numba
        lm = np.arange(self.image.shape[1]).astype(self.image.dtype) * self.lm_scale + self.lm_bias
        lm2 = lm * lm
        n = np.sqrt(1 - (lm2[:, np.newaxis] + lm2[np.newaxis, :]))[np.newaxis, ...]
        w_correct = np.exp(-2j * math.pi * self.w * (n - 1))
        kernel = np.outer(self.kernel1d, self.kernel1d)[np.newaxis, ...]
        self.layer[:] = self.image / (kernel * n) * w_correct
        self.grid[:] = np.fft.fftshift(
            np.fft.fft2(
                np.fft.ifftshift(self.layer, axes=(1, 2)),
                axes=(1, 2)),
            axes=(1, 2))
